# inside your module, e.g. utils.py
import os
import sys

import ast
import textwrap
from pathlib import Path


def _is_plain_check_run(test: ast.AST) -> bool:
    """True only for a bare call to check_run() as the entire if-test."""
    if not isinstance(test, ast.Call) or test.args or test.keywords:
        return False
    f = test.func
    return (
        isinstance(f, ast.Attribute) and f.attr == "cache_run"
    ) or (
        isinstance(f, ast.Name) and f.id == "cache_run"
    )

def get_run_block_str(source: str) -> str:
    """
    Return all source text before the first `if check_run():` / `if *.check_run():`.
    Returns None if no matching if-statement exists.
    """
    tree = ast.parse(source)
    candidates = [n for n in ast.walk(tree) if isinstance(n, ast.If) and _is_plain_check_run(n.test)]
    if not candidates:
        return None
    first = min(candidates, key=lambda n: n.lineno)
    lines = source.splitlines(keepends=True)
    return "".join(lines[: first.lineno - 1])


def get_build_block_str(source: str) -> str | None:
    """Return the code string inside the first `if *.checkrun(...):` block, or None."""
    tree = ast.parse(source)
    for node in ast.walk(tree):
        if isinstance(node, ast.If) and isinstance(node.test, ast.Call):
            f = node.test.func
            called = (isinstance(f, ast.Attribute) and f.attr == "cache_build") or \
                     (isinstance(f, ast.Name) and f.id == "cache_build")
            if called and node.body:
                start = node.body[0].lineno - 1           # 0-based start line
                end = node.body[-1].end_lineno            # 1-based end line (inclusive)
                lines = source.splitlines()
                block = "\n".join(lines[start:end])
                return textwrap.dedent(block)
    return None

def entry_script_path():
    main = sys.modules.get("__main__")
    # Most normal runs: python path/to/app.py
    if main and hasattr(main, "__file__"):
        return os.path.abspath(main.__file__)
    # Fallbacks (e.g. python -m pkg.mod)
    if sys.argv and sys.argv[0] not in ("", "-c"):
        return os.path.abspath(sys.argv[0])
    # Interactive sessions, notebooks, or embedded interpreters
    return None


def get_build_section() -> str:
    """ Returns the string section inside the check_build() if statement"""
    name = entry_script_path()

    lines = get_build_block_str(Path(name).read_text())
    return lines

def get_run_section() -> str:
    """Return sthe string section before the check_run() if statement

    Returns:
        str: _description_
    """
    name = entry_script_path()

    lines = get_run_block_str(Path(name).read_text())
    return lines


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import numpy as np
from loguru import logger
from typing import TypeVar, Generic, Any, List, Union, Dict, Generator
from collections import defaultdict

T = TypeVar("T")
M = TypeVar("M")

def assemble_nd_data(
    data_in: List[Union[float, np.ndarray]],
    vars_in: List[Dict[str, float]],
    axes: Dict[str, List[float]],
) -> np.ndarray:
    """
    Assemble a flat list of data entries into an N-dimensional array
    based on provided axis definitions.

    Parameters
    ----------
    data_in : List[float or np.ndarray]
        Flat list of data entries. Scalars are treated as shape ().
    vars_in : List[Dict[str, float]]
        List of dictionaries mapping variable names to their value for each data entry.
        Length must match data_in.
    axes : Dict[str, List[float]]
        Mapping from variable name to list of allowed values along that axis.
        If 'freq' is a key, its axis will be placed last.

    Returns
    -------
    np.ndarray
        An array of shape (*axis_lengths, *shp_out), where axis_lengths correspond
        to lengths of each axis in `axes` (with 'freq' last if present) and
        shp_out is the shape of the first ndarray in data_in or () if all scalars.
    """
    # Determine axis ordering
    axis_names = sorted(list(axes.keys()))
    if 'freq' in axis_names:
        axis_names.append(axis_names.pop(axis_names.index('freq')))
    axis_lengths = [len(axes[name]) for name in axis_names]

    # Determine output shape suffix
    # Assume all array entries share the same shape
    first = next((d for d in data_in if isinstance(d, np.ndarray)), None)
    shp_out = first.shape if isinstance(first, np.ndarray) else ()

    # Initialize output array
    out_shape = tuple(axis_lengths) + shp_out
    out = np.empty(out_shape, dtype=first.dtype if isinstance(first, np.ndarray) else float)

    # Fill in data
    for entry, var_map in zip(data_in, vars_in):
        # build index for each axis
        idx = tuple(
            axes[name].index(var_map[name])
            for name in axis_names
        )
        if isinstance(entry, np.ndarray):
            out[idx + tuple(slice(None) for _ in shp_out)] = entry
        else:
            out[idx] = entry

    return out

def generate_ndim(
    outer_data: dict[str, list[float]],
    inner_data: list[float],
    outer_labels: tuple[str, ...]
) -> tuple[np.ndarray,...]:
    """
    Generates an N-dimensional grid of values from flattened data, and returns each axis array plus the grid.

    Parameters
    ----------
    outer_data : dict of {label: flat list of coordinates}
        Each key corresponds to one axis label, and the list contains coordinate values for each point.
    inner_data : list of float
        Flattened list of data values corresponding to each set of coordinates.
    outer_labels : tuple of str
        Order of axes (keys of outer_data) which defines the dimension order in the output array.

    Returns
    -------
    *axes : np.ndarray
        One 1D array for each axis, containing the sorted unique coordinates for that dimension, 
        in the order specified by outer_labels.
    grid : np.ndarray
        N-dimensional array of shape (n1, n2, ..., nN), where ni is the number of unique
        values along the i-th axis. Missing points are filled with np.nan.
    """
    # Convert inner data to numpy array
    values = np.asarray(inner_data)

    # Determine unique sorted coordinates for each axis
    axes = [np.unique(np.asarray(outer_data[label])) for label in outer_labels]
    grid_shape = tuple(axis.size for axis in axes)

    # Initialize grid with NaNs
    grid = np.full(grid_shape, np.nan, dtype=values.dtype)

    # Build coordinate arrays for each axis
    coords = [np.asarray(outer_data[label]) for label in outer_labels]

    # Map coordinates to indices in the grid for each axis
    idxs = [np.searchsorted(axes[i], coords[i]) for i in range(len(axes))]

    # Assign values into the grid
    grid[tuple(idxs)] = values

    # Return each axis array followed by the grid
    return tuple(axes) + (grid,)


class DataEntry:
    
    def __init__(self, vars: dict[str, float]):
        self.vars: dict[str, float] = vars
        self.data: dict[str, Any] = dict()
    
    def print(self) -> None:
        """Print the content of the DataEntry object"""
        for key, value in self.data.items():
            print(f'    {key} = {value}')

    def values(self) -> list[Any]:
        """ Return all values stored in the DataEntry"""
        return list(self.data.values())
    
    def keys(self) -> list[str]:
        """ Return all names of data stored in the DataEntry"""
        return list(self.data.keys())
    
    def items(self) -> list[tuple[str, Any]]:
        """ Returns a list of all key: value pairs of the DataEntry."""
        return list(self.data.items())
    
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, dict):
            return False
        allkeys = set(list(self.vars.keys()) + list(other.keys()))
        return all(self.vars[key]==other[key] for key in allkeys)
    
    def _dist(self, other: dict[str, float]) -> float:
        return sum([(abs(self.vars.get(key,1e20)-other[key])/(other[key]+1e-12)) for key in other.keys()])
    
    def __getitem__(self, key) -> Any:
        return self.data[key]
    
    def __setitem__(self, key: str, value: Any) -> None:
        self.data[key] = value


class DataContainer:
    """The DataContainer class is a generalized class to store data for a set of parameter sweeps"""
    def __init__(self):
        self.entries: list[DataEntry] = []
        self.stock: DataEntry = DataEntry({})
        

    def print(self) -> None:
        """ Print an overview of all data in the DataContainer"""
        self.stock.print()
        for entry in self.entries:
            entry.print()

    def new(self, **vars: float) -> DataEntry:
        """Create a new entry in the DataContainer for the given value setting"""
        entry = DataEntry(vars)
        self.entries.append(entry)
        return entry
    
    def iterate(self) -> Generator[tuple[dict[str, float], dict[str, Any]], None, None]:
        for entry in self.entries:
            yield entry.vars, entry.data

    @property
    def first(self) -> DataEntry:
        """Returns the first added entry"""
        return self.entries[0]
    
    @property
    def last(self) -> DataEntry:
        """Returns the last added entry"""
        return self.entries[-1]
    
    @property
    def default(self) -> DataEntry:
        """Returns the default DataEntry which is either the last from the parameter sweep or the general one in case of no parameter sweep."""
        if not self.entries:
            return self.stock
        else:
            return self.last
    
    def index(self, index: int) -> DataEntry:
        """Returns the last added entry"""
        return self.entries[index]
    
    def select(self, **vars: float) -> DataEntry | None:
        """Returns the data entry corresponding to the provided parametric sweep set"""
        for entry in self.entries:
            if entry==vars:
                return entry
        return None
    
    def find(self, **vars: float) -> DataEntry:
        """Returns the DataEntry closest to the provided parametric sweep setting."""
        return sorted([(entry, entry._dist(vars)) for entry in self.entries], key=lambda x: x[1])[0][0]
    
    def __getitem__(self, key: str) -> DataEntry:
        """Returns the requested item from the default DataEntry"""
        return self.default[key]
        
    def __setitem__(self, key: str, value: Any) -> None:
        """Writes a value to the requested default DataEntry"""
        self.default[key] = value
    

class BaseDataset(Generic[T,M]):
    def __init__(self, datatype: type[T], matrixtype: type[M], scalar: bool):
        self._datatype: type[T] = datatype
        self._matrixtype: type[M] = matrixtype
        self._variables: list[dict[str, float]] = []
        self._data_entries: list[T] = []
        self._scalar: bool = scalar
        
        self._gritted: bool | None = None
        self._axes: dict[str, np.ndarray]| None = None
        self._ax_ids: dict[str, int]| None = None
        self._ids: np.ndarray| None = None
        self._gridobj: M | None = None

        self._data: dict[str, Any] = dict()
    
    def __getitem__(self, index: int) -> T:
        return self._data_entries[index]

    @property
    def _fields(self) -> list[str]:
        return self._datatype._fields # type: ignore
    
    @property
    def _copy(self) -> list[str]:
        return self._datatype._copy # type: ignore
    
    def store(self, key: str, value: Any) -> None:
        """Stores a variable with some value in the provided key. 
        Make sure that all values passed are picklable.

        Args:
            key (str): The name for the data entry
            value (Any): The value of the data entry
        """
        self._data[key] = value

    def load(self, key: str) -> Any | None:
        """Returns the data entry for a given key

        Args:
            key (str): The name of the data entry

        Returns:
            Any: The value of the data entry
        """
        return self._data.get(key, None)
    
    def get_entry(self, index: int) -> T:
        """Returns the physics dataset for  the given index

        Args:
            index (int): The index of the solution

        Returns:
            T: The Physics dataset
        """
        return self._data_entries[index]

    def select(self, **variables: float) -> T:
        """Returns the first physics dataset that satisfies the variable assignemnt.

        Returns:
            T: The physics dataset
        """
        index = next(
            i
            for i, var_map in enumerate(self._variables)
            if all(var_map.get(k) == v for k, v in variables.items())
        )
        return self.get_entry(index)
    
    def filter(self, **variables: float) -> list[T]:
        """Returns a list of all physics datasets that are valid for the given variable assignment

        Returns:
            list[T]: A list of all matching datasets
        """
        output = []
        for i, var_map in enumerate(self._variables):
            if all(var_map.get(k) == v for k, v in variables.items()):
                output.append(self.get_entry(i))
        return output
    
    def find(self, **variables: float) -> T:
        """Returns the physics dataset that is closest to the constraint given by the variables.

        Returns:
            T: The physics dataset.
        """
        output = []
        for i, var_map in enumerate(self._variables):
            error = sum([abs(var_map.get(k, 1e30) - v) for k, v in variables.items()])
            output.append((i,error))
        selection_id = sorted(output, key=lambda x:x[1])[0][0]
        entry = self.get_entry(selection_id)
        variables = ', '.join([f'{key}={value}' for key,value in self._variables[selection_id].items()])
        logger.info(f'Selected entry: {variables}')
        return entry

    def axis(self, name: str) -> np.ndarray:
        """Returns a sorted list of all variables for the given name

        Args:
            name (str): The name of the variable axis

        Returns:
            np.ndarray: A sorted list of all unique values.
        """
        return np.sort(np.unique(np.array([var[name] for var in self._variables])))

    def new(self, **vars: float) -> T:
        """Creates a new dataset

        Returns:
            T: The physics dataset object
        """
        self._variables.append(vars)
        new_entry = self._datatype()
        self._data_entries.append(new_entry)
        return new_entry
    
    def _grid_axes(self) -> bool:
        """This method attepmts to create a gritted version of the scalar dataset. It may fail
        if the data in the dataset cannot be cast into a gridded structure.

        Returns:
            None
        """
        logger.debug('Attempting to grid simulation data')
        variables = defaultdict(set)
        for var in self._variables:
            for key, value in var.items():
                variables[key].add(value)
                
        N_entries = len(self._variables)
        N_prod = 1
        N_dim = len(variables)
        
        for key, val_list in variables.items():
            N_prod *= len(val_list)
        
        if N_entries == N_prod:
            logger.debug('Multi-dimensional grid found!')
            self._gritted = True
        else:
            logger.debug('Multi-dimensional grid not found')
            self._gritted = False
            return False
        
        self._axes = dict()
        self._ax_ids = dict()
        
        revax = dict()
        i = 0
        
        for key, val_set in variables.items():
            self._axes[key] = np.sort(np.array(list(val_set)))
            self._ax_ids[key] = i
            revax[i] = key
            i += 1
            
        axlist = []

        for idim in range(N_dim):
            axlist.append(self._axes[revax[idim]])
        
        indices = np.arange(N_entries)
        Ndimlist = [len(dim) for dim in axlist]
        self._ids = indices.reshape(Ndimlist)

        obj = self._matrixtype()

        axes_list = {key: list(self._axes[key]) for key in self._axes.keys()}
        for field in self._fields:
            data_field = [self._data_entries[i].__dict__[field] for i in range(N_entries)]
            data_set = assemble_nd_data(data_field, self._variables, axes_list)
            obj.__setattr__(field, data_set)
        for copyfield in self._copy:
            
            obj.__setattr__(copyfield, self._data_entries[0].__dict__[copyfield])
        self._gridobj = obj

        return True
    
    @property
    def grid(self) -> M:
        """Returns the gridded version of the scalar dataset.

        Raises:
            ValueError: _description_

        Returns:
            M: The gritted physics dataset
        """
        if self._gritted is None:
            self._grid_axes()

        if self._gritted is False:
            logger.error('The dataset cannot be cast to a structured grid.')
            raise ValueError('Data not in regular grid')
        
        return self._gridobj


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from ...mesher import Mesher
from ...material import Material
from ...mesh3d import Mesh3D
from ...coord import Line
from ...geometry import GeoSurface
from ...elements.femdata import FEMBasis
from ...elements.nedelec2 import Nedelec2
from ...solver import DEFAULT_ROUTINE, SolveRoutine
from ...system import called_from_main_function
from ...selection import FaceSelection
from ...settings import Settings
from .microwave_bc import MWBoundaryConditionSet, PEC, ModalPort, LumpedPort, PortBC
from .microwave_data import MWData
from .assembly.assembler import Assembler
from .port_functions import compute_avg_power_flux
from .simjob import SimJob

from concurrent.futures import ThreadPoolExecutor
from loguru import logger
from typing import Callable, Literal
import multiprocessing as mp
from cmath import sqrt as csqrt

import numpy as np
import threading
import time

class SimulationError(Exception):
    pass

def run_job_multi(job: SimJob) -> SimJob:
    """The job launcher for Multi-Processing environements

    Args:
        job (SimJob): The Simulation Job

    Returns:
        SimJob: The solved SimJob
    """
    routine = DEFAULT_ROUTINE._configure_routine('MP')
    for A, b, ids, reuse, aux in job.iter_Ab():
        solution, report = routine.solve(A, b, ids, reuse, id=job.id)
        report.add(**aux)
        job.submit_solution(solution, report)
    return job

def _init_worker():
    nr = int(mp.current_process().name.split('-')[1])
    DEFAULT_ROUTINE._configure_routine(proc_nr=nr)

def _dimstring(data: list[float] | np.ndarray) -> str:
    """A String formatter for dimensions in millimeters

    Args:
        data (list[float]): The list of floating point dimensions

    Returns:
        str: The formatted string
    """
    return '(' + ', '.join([f'{x*1000:.1f}mm' for x in data]) + ')'

def shortest_path(xyz1: np.ndarray, xyz2: np.ndarray, Npts: int) -> np.ndarray:
    """
    Finds the pair of points (one from xyz1, one from xyz2) that are closest in Euclidean distance,
    and returns a (3, Npts) array of points linearly interpolating between them.

    Parameters:
    xyz1 : np.ndarray of shape (3, N1)
    xyz2 : np.ndarray of shape (3, N2)
    Npts : int, number of points in the output path

    Returns:
    np.ndarray of shape (3, Npts)
    """
    # Compute pairwise distances (N1 x N2)
    diffs = xyz1[:, :, np.newaxis] - xyz2[:, np.newaxis, :]
    dists = np.linalg.norm(diffs, axis=0)  # shape (N1, N2)

    # Find indices of closest pair
    i1, i2 = np.unravel_index(np.argmin(dists), dists.shape)
    p1 = xyz1[:, i1]
    p2 = xyz2[:, i2]

    # Interpolate linearly between p1 and p2
    t = np.linspace(0, 1, Npts)
    path = (1 - t) * p1[:, np.newaxis] + t * p2[:, np.newaxis]

    return path
    
class Microwave3D:
    """The Electrodynamics time harmonic physics class.

    This class contains all physics dependent features to perform EM simuation in the time-harmonic
    formulation.

    """
    def __init__(self, mesher: Mesher, settings: Settings, mwdata: MWData, order: int = 2):
        self.frequencies: list[float] = []
        self.current_frequency = 0
        self.order: int = order
        self.resolution: float = 1
        self._settings: Settings = settings
        self.mesher: Mesher = mesher
        self.mesh: Mesh3D = Mesh3D(self.mesher)

        self.assembler: Assembler = Assembler(self._settings)
        self.bc: MWBoundaryConditionSet = MWBoundaryConditionSet(None)
        self.basis: Nedelec2 | None = None
        self.solveroutine: SolveRoutine = DEFAULT_ROUTINE
        self.cache_matrices: bool = True

        ## States
        self._bc_initialized: bool = False
        self.data: MWData = mwdata

        ## Data
        self._params: dict[str, float] = dict()
        self._simstart: float = 0.0
        self._simend: float = 0.0

        self.set_order(order)

    def reset_data(self):
        self.data = MWData()

    def reset(self):
        self.bc.reset()
        self.basis: FEMBasis = None
        self.bc = MWBoundaryConditionSet(None)
        self.solveroutine.reset()

    def set_order(self, order: int) -> None:
        """Sets the order of the basis functions used. Currently only supports second order.

        Args:
            order (int): The order to use.

        Raises:
            ValueError: An error if a wrong order is used.
        """
        if order not in (2,):
            raise ValueError(f'Order {order} not supported. Only order-2 allowed.')
        
        self.order = order
        self.resolution = {1: 0.15, 2: 0.3}[order]

    @property
    def nports(self) -> int:
        """The number of ports in the physics.

        Returns:
            int: The number of ports
        """
        return self.bc.count(PortBC)
    
    def ports(self) -> list[PortBC]:
        """A list of all port boundary conditions.

        Returns:
            list[PortBC]: A list of all port boundary conditions
        """
        return sorted(self.bc.oftype(PortBC), key=lambda x: x.number) # type: ignore
    
    
    def _initialize_bcs(self, surfaces: list[GeoSurface]) -> None:
        """Initializes the boundary conditions to set PEC as all exterior boundaries.
        """
        logger.debug('Initializing boundary conditions.')

        tags = self.mesher.domain_boundary_face_tags
        
        # Assigning surface impedance boundary condition
        if self._settings.mw_2dbc:
            for surf in surfaces:
                if surf.material.cond.scalar(1e9) > self._settings.mw_2dbc_peclim:
                    logger.debug(f'Assinging PEC to {surf}')
                    self.bc.PEC(surf)
                elif surf.material.cond.scalar(1e9) > self._settings.mw_2dbc_lim:
                    self.bc.SurfaceImpedance(surf, surf.material)
                
        
        tags = [tag for tag in tags if tag not in self.bc.assigned(2)]
        
        self.bc.PEC(FaceSelection(tags))
        
        logger.info(f'Adding PEC boundary condition with tags {tags}.')
        if self.mesher.periodic_cell is not None:
            self.mesher.periodic_cell.generate_bcs()
            for bc in self.mesher.periodic_cell.bcs:
                self.bc.assign(bc)

    def set_frequency(self, frequency: float | list[float] | np.ndarray ) -> None:
        """Define the frequencies for the frequency sweep

        Args:
            frequency (float | list[float] | np.ndarray): The frequency points.
        """
        logger.info(f'Setting frequency as {frequency}Hz.')
        if isinstance(frequency, (tuple, list, np.ndarray)):
            self.frequencies = list(frequency)
        else:
            self.frequencies = [frequency]

        self.mesher.max_size = self.resolution * 299792458 / max(self.frequencies)
        self.mesher.min_size = 0.1 * self.mesher.max_size

        logger.debug(f'Setting global mesh size range to: {self.mesher.min_size*1000:.3f}mm - {self.mesher.max_size*1000:.3f}mm')
    
    def set_frequency_range(self, fmin: float, fmax: float, Npoints: int) -> None:
        """Set the frequency range using the np.linspace syntax

        Args:
            fmin (float): The starting frequency
            fmax (float): The ending frequency
            Npoints (int): The number of points
        """
        self.set_frequency(np.linspace(fmin, fmax, Npoints))
    
    def fdense(self, Npoints: int) -> np.ndarray:
        """Return a resampled version of the current frequency range

        Args:
            Npoints (int): The new number of points

        Returns:
            np.ndarray: The new frequency axis
        """
        if len(self.frequencies) == 1:
            raise ValueError('Only 1 frequency point known. At least two need to be defined.')
        fmin = min(self.frequencies)
        fmax = max(self.frequencies)
        return np.linspace(fmin, fmax, Npoints)
    
    def set_resolution(self, resolution: float) -> None:
        """Define the simulation resolution as the fraction of the wavelength.

        To define the wavelength as ¼λ, call .set_resolution(0.25)

        Args:
            resolution (float): The desired wavelength fraction.
            
        """
        if resolution > 0.5:
            logger.warning('Resolutions greater than 0.5 cannot yield accurate results, capping resolution to 0.4')
            resolution = 0.4
        elif resolution > 0.334:
            logger.warning('A resolution greater than 0.33 may cause accuracy issues.')
        
        self.resolution = resolution
        logger.trace(f'Resolution set to {self.resolution}')

    def set_conductivity_limit(self, condutivity: float) -> None:
        """Sets the limit of a material conductivity value beyond which
        the assembler considers it PEC. By default this value is
        set to 1·10⁷S/m which means copper conductivity is ignored.

        Args:
            condutivity (float): The conductivity level in S/m
        """
        if condutivity < 0:
            raise ValueError('Conductivity values must be above 0. Ignoring assignment')

        self.assembler.settings.mw_2dbc_peclim = condutivity
        self.assembler.settings.mw_3d_peclim = condutivity
        logger.trace(f'Set conductivity limit to {condutivity} S/m')
    
    def get_discretizer(self) -> Callable:
        """Returns a discretizer function that defines the maximum mesh size.

        Returns:
            Callable: The discretizer function
        """
        def disc(material: Material):
            return 299792458/(max(self.frequencies) * np.real(material.neff(max(self.frequencies))))
        return disc
    
    def _initialize_field(self):
        """Initializes the physics basis to the correct FEMBasis object.
        
        Currently it defaults to Nedelec2. Mixed basis are used for modal analysis. 
        This function does not have to be called by the user. Its automatically invoked.
        """
        if self.basis is not None:
            return
        if self.order == 1:
            raise NotImplementedError('Nedelec 1 is currently not supported')
            from ...elements import Nedelec1
            self.basis = Nedelec1(self.mesh)
        elif self.order == 2:
            from ...elements.nedelec2 import Nedelec2
            self.basis = Nedelec2(self.mesh)

    def _initialize_bc_data(self):
        ''' Initializes auxilliary required boundary condition information before running simulations.
        '''
        logger.debug('Initializing boundary conditions')
        self.bc.cleanup()
        for port in self.bc.oftype(LumpedPort):
            self.define_lumped_port_integration_points(port)
    
    def _check_physics(self) -> None:
        """ Executes a physics check before a simulation can be run."""
        if not self.bc._is_excited():
            raise SimulationError('The simulation has no boundary conditions that insert energy. Make sure to include at least one Port into your simulation.')
    
    def define_lumped_port_integration_points(self, port: LumpedPort) -> None:
        """Sets the integration points on Lumped Port objects for voltage integration

        Args:
            port (LumpedPort): The LumpedPort object

        Raises:
            SimulationError: An error if there are no nodes associated with the port.
        """
        if len(port.vintline) > 0:
            return
        logger.debug(' - Finding Lumped Port integration points')
        field_axis = port.Vdirection.np

        points = self.mesh.get_nodes(port.tags)

        if points.size==0:
            raise SimulationError(f'The lumped port {port} has no nodes associated with it')
        
        xs = self.mesh.nodes[0,points]
        ys = self.mesh.nodes[1,points]
        zs = self.mesh.nodes[2,points]

        dotprod = xs*field_axis[0] + ys*field_axis[1] + zs*field_axis[2]

        start_id = np.argwhere(dotprod == np.min(dotprod)).flatten()
        
        xs = xs[start_id]
        ys = ys[start_id]
        zs = zs[start_id]
        

        for x,y,z in zip(xs, ys, zs):
            start = np.array([x,y,z])
            end = start + port.Vdirection.np*port.height
            port.vintline.append(Line.from_points(start, end, 21))
            logger.trace(f' - Port[{port.port_number}] integration line {start} -> {end}.')
        
        port.v_integration = True
    
    def _compute_integration_line(self, group1: list[int], group2: list[int]) -> tuple[np.ndarray, np.ndarray]:
        """Computes an integration line for two node island groups by finding the closest two nodes.
        
        This method is used for the modal TEM analysis to find an appropriate voltage integration path
        by looking for the two closest points for the two conductor islands that where discovered.

        Currently it defaults to 11 integration line points.

        Args:
            group1 (list[int]): The first island node group
            group2 (list[int]): The second island node group

        Returns:
            centers (np.ndarray): The center points of the line segments
            dls (np.ndarray): The delta-path vectors for each line segment.
        """
        nodes1 = self.mesh.nodes[:,group1]
        nodes2 = self.mesh.nodes[:,group2]
        path = shortest_path(nodes1, nodes2, 21)
        centres = (path[:,1:] + path[:,:-1])/2
        dls = path[:,1:] - path[:,:-1]
        return centres, dls

    def _find_tem_conductors(self, port: ModalPort, sigtri: np.ndarray) -> tuple[list[int], list[int]]:
        ''' Returns two lists of global node indices corresponding to the TEM port conductors.
        
        This method is invoked during modal analysis with TEM modes. It looks at all edges
        exterior to the boundary face triangulation and finds two small subsets of nodes that
        lie on different exterior boundaries of the boundary face.

        Args:
            port (ModalPort): The modal port object.
            
        Returns:
            list[int]: A list of node integers of island 1.
            list[int]: A list of node integers of island 2.
        '''
        if self.basis is None:
            raise ValueError('The field basis is not yet defined.')

        logger.debug(' - Finding PEC TEM conductors')
        pecs: list[PEC] = self.bc.get_conductors() # type: ignore
        mesh = self.mesh

        # Process all PEC Boundary Conditions
        pec_edges = []
        for pec in pecs:
            face_tags = pec.tags
            tri_ids = mesh.get_triangles(face_tags)
            edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())
            pec_edges.extend(edge_ids)
        
        # Process conductivity
        for itri in mesh.get_triangles(port.tags):
            if sigtri[itri] > 1e6:
                edge_ids = list(mesh.tri_to_edge[:,itri].flatten())
                pec_edges.extend(edge_ids)

        pec_edges = list(set(pec_edges))
        
        tri_ids = mesh.get_triangles(port.tags)
        edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())
        
        pec_port = np.array([i for i in pec_edges if i in set(edge_ids)])
        
        pec_islands = mesh.find_edge_groups(pec_port)

        logger.debug(f' - Found {len(pec_islands)} PEC islands.')

        if len(pec_islands) != 2:
            raise ValueError(f' - Found {len(pec_islands)} PEC islands. Expected 2.')
        
        groups = []
        for island in pec_islands:
            group = set()
            for edge in island:
                group.add(mesh.edges[0,edge])
                group.add(mesh.edges[1,edge])
            groups.append(sorted(list(group)))
        
        group1 = groups[0]
        group2 = groups[1]

        return group1, group2
    
    def _compute_modes(self, freq: float):
        """Compute the modal port modes for a given frequency. Used internally by the frequency domain study.

        Args:
            freq (float): The simulation frequency
        """
        for bc in self.bc.oftype(ModalPort):
            
            # If there is a port mode (at least one) and the port does not have mixed materials. No new analysis is needed
            if not bc.mixed_materials and bc.initialized:
                continue
            
            self.modal_analysis(bc, 1, False, bc.TEM, freq=freq)

    def modal_analysis(self, 
                       port: ModalPort, 
                       nmodes: int = 6, 
                       direct: bool = True,
                       TEM: bool = False,
                       target_kz = None,
                       target_neff = None,
                       freq: float | None = None) -> None:
        ''' Execute a modal analysis on a given ModalPort boundary condition.
        
        Parameters:
        -----------
            port : ModalPort
                The port object to execute the analysis for.
            direct : bool
                Whether to use the direct solver (LAPACK) if True. Otherwise it uses the iterative
                ARPACK solver. The ARPACK solver required an estimate for the propagation constant and is faster
                for a large number of Degrees of Freedom.
            TEM : bool = True
                Whether to estimate the propagation constant assuming its a TEM transmisison line.
            target_k0 : float
                The expected propagation constant to find a mode for (direct = False).
            target_neff : float
                The expected effective mode index defined as kz/k0 (1.0 = free space, <1 = TE/TM, >1=slow wavees)
            freq : float = None
                The desired frequency at which the mode is solved. If None then it uses the lowest frequency of the provided range.
        '''
        T0 = time.time()
        logger.info(f'Starting Mode Analysis for port {port}.')
        
        if self.bc._initialized is False:
            raise SimulationError('Cannot run a modal analysis because no boundary conditions have been assigned.')
        
        self._initialize_field()
        self._initialize_bc_data()

        if self.basis is None:
            raise SimulationError('Cannot proceed, the current basis class is undefined.')

        logger.debug(' - retreiving material properties.')
        
        if freq is None:
            freq = self.frequencies[0]
        
        materials = self.mesh._get_material_assignment(self.mesher.volumes)

        ertet = np.zeros((3,3,self.mesh.n_tets), dtype=np.complex128)
        tandtet = np.zeros((3,3,self.mesh.n_tets), dtype=np.complex128)
        urtet = np.zeros((3,3,self.mesh.n_tets), dtype=np.complex128)
        condtet = np.zeros((3,3,self.mesh.n_tets), dtype=np.complex128)
        
        for mat in materials:
            ertet = mat.er(freq, ertet)
            tandtet = mat.tand(freq, tandtet)
            urtet = mat.ur(freq, urtet)
            condtet = mat.cond(freq, condtet)
        
        ertet = ertet * (1-1j*tandtet)
        
        er = np.zeros((3,3,self.mesh.n_tris,), dtype=np.complex128)
        ur = np.zeros((3,3,self.mesh.n_tris,), dtype=np.complex128)
        cond = np.zeros((self.mesh.n_tris,), dtype=np.complex128)

        for itri in range(self.mesh.n_tris):
            itet = self.mesh.tri_to_tet[0,itri]
            er[:,:,itri] = ertet[:,:,itet]
            ur[:,:,itri] = urtet[:,:,itet]
            cond[itri] = condtet[0,0,itet]

        itri_port = self.mesh.get_triangles(port.tags)

        ermean = np.mean(er[er>0].flatten()[itri_port])
        urmean = np.mean(ur[ur>0].flatten()[itri_port])
        ermax = np.max(er[:,:,itri_port].flatten())
        urmax = np.max(ur[:,:,itri_port].flatten())

        k0 = 2*np.pi*freq/299792458
        
        logger.debug(f' - mean(max): εr = {ermean:.2f}({ermax:.2f}), μr = {urmean:.2f}({urmax:.2f})')
        
        Amatrix, Bmatrix, solve_ids, nlf = self.assembler.assemble_bma_matrices(self.basis, er, ur, cond, k0, port, self.bc)
        
        logger.debug(f'Total of {Amatrix.shape[0]} Degrees of freedom.')
        logger.debug(f'Applied frequency: {freq/1e9:.2f}GHz')
        logger.debug(f'K0 = {k0} rad/m')

        F = -1

        if target_neff is not None:
            target_kz = k0*target_neff
        
        if target_kz is None:
            if TEM:
                target_kz = ermean*urmean*1.1*k0
            else:
                
                target_kz = ermean*urmean*0.7*k0
                
        logger.debug(f'Solving for {solve_ids.shape[0]} degrees of freedom.')

        eigen_values, eigen_modes, report = self.solveroutine.eig_boundary(Amatrix, Bmatrix, solve_ids, nmodes, direct, target_kz, sign=-1)
        
        logger.debug(f'Eigenvalues: {np.sqrt(F*eigen_values)} rad/m')

        port._er = er
        port._ur = ur

        nmodes_found = eigen_values.shape[0]

        for i in range(nmodes_found):
            
            Emode = np.zeros((nlf.n_field,), dtype=np.complex128)
            eigenmode = eigen_modes[:,i]
            Emode[solve_ids] = np.squeeze(eigenmode)
            Emode = Emode * np.exp(-1j*np.angle(np.max(Emode)))

            beta_base = np.emath.sqrt(-eigen_values[i])
            beta = min(k0*np.sqrt(ermax*urmax), beta_base)

            residuals = -1

            portfE = nlf.interpolate_Ef(Emode)
            portfH = nlf.interpolate_Hf(Emode, k0, ur, beta)

            P = compute_avg_power_flux(nlf, Emode, k0, ur, beta)

            mode = port.add_mode(Emode, portfE, portfH, beta, k0, residuals, TEM=TEM, freq=freq)
            if mode is None:
                continue
            
            Efxy = Emode[:nlf.n_xy]
            Efz = Emode[nlf.n_xy:]
            Ez = np.max(np.abs(Efz))
            Exy = np.max(np.abs(Efxy))
            
            if Ez/Exy < 1e-1 and not TEM:
                logger.debug('Low Ez/Et ratio detected, assuming TE mode')
                mode.modetype = 'TE'
            elif Ez/Exy > 1e-1 and not TEM:
                logger.debug('High Ez/Et ratio detected, assuming TM mode')
                mode.modetype = 'TM'
            elif TEM:
                G1, G2 = self._find_tem_conductors(port, sigtri=cond)
                cs, dls = self._compute_integration_line(G1,G2)
                mode.modetype = 'TEM'
                Ex, Ey, Ez = portfE(cs[0,:], cs[1,:], cs[2,:])
                voltage = np.sum(Ex*dls[0,:] + Ey*dls[1,:] + Ez*dls[2,:])
                mode.Z0 = abs(voltage**2/(2*P))
                logger.debug(f'Port Z0 = {mode.Z0}')

            mode.set_power(P*port._qmode(k0)**2)
        
        port.sort_modes()

        logger.info(f'Total of {port.nmodes} found')

        T2 = time.time()    
        logger.info(f'Elapsed time = {(T2-T0):.2f} seconds.')
        return None
    
    def run_sweep(self, 
                parallel: bool = False,
                n_workers: int = 2, 
                harddisc_threshold: int | None = None,
                harddisc_path: str = 'EMergeSparse',
                frequency_groups: int = -1,
                multi_processing: bool = False,
                automatic_modal_analysis: bool = True) -> MWData:
        """Executes a frequency domain study

        The study is distributed over "n_workers" workers.
        As optional parameter you may set a harddisc_threshold as integer. This determines the maximum
        number of degrees of freedom before which the jobs will be cahced to the harddisk. The
        path that will be used to cache the sparse matrices can be specified.
        Additionally the term frequency_groups may be specified. This number will define in how
        many groups the matrices will be pre-computed before they are send to workers. This can minimize
        the total amound of RAM memory used. For example with 11 frequencies in gruops of 4, the following
        frequency indices will be precomputed and then solved: [[1,2,3,4],[5,6,7,8],[9,10,11]]

        Args:
            n_workers (int, optional): The number of workers. Defaults to 2.
            harddisc_threshold (int, optional): The number of DOF limit. Defaults to None.
            harddisc_path (str, optional): The cached matrix path name. Defaults to 'EMergeSparse'.
            frequency_groups (int, optional): The number of frequency points in a solve group. Defaults to -1.
            automatic_modal_analysis (bool, optional): Automatically compute port modes. Defaults to False.
            multi_processing (bool, optional): Whether to use multiprocessing instead of multi-threaded (slower on most machines).

        Raises:
            SimulationError: An error associated witha a problem during the simulation.

        Returns:
            MWSimData: The dataset.
        """
        
        self._simstart = time.time()
        if self.bc._initialized is False:
            raise SimulationError('Cannot run a modal analysis because no boundary conditions have been assigned.')
        
        self._initialize_field()
        self._initialize_bc_data()
        self._check_physics()
        
        if self.basis is None:
            raise SimulationError('Cannot proceed, the simulation basis class is undefined.')

        materials = self.mesh._get_material_assignment(self.mesher.volumes)

        ### Does this move
        logger.debug('Initializing frequency domain sweep.')
        
        #### Port settings
        all_ports = self.bc.oftype(PortBC)

        ##### FOR PORT SWEEP SET ALL ACTIVE TO FALSE. THIS SHOULD BE FIXED LATER
        ### COMPUTE WHICH TETS ARE CONNECTED TO PORT INDICES

        for port in all_ports:
            port.active=False
            

        logger.info(f'Pre-assembling matrices of {len(self.frequencies)} frequency points.')

        thread_local = None
        if parallel:
            # Thread-local storage for per-thread resources
            thread_local = threading.local()

        ## DEFINE SOLVE FUNCTIONS
        def get_routine():
            if not hasattr(thread_local, "routine"):
                worker_nr = int(threading.current_thread().name.split('_')[1])+1
                thread_local.routine = self.solveroutine.duplicate()._configure_routine('MT', thread_nr=worker_nr)
            return thread_local.routine

        def run_job(job: SimJob):
            routine = get_routine()
            for A, b, ids, reuse, aux in job.iter_Ab():
                solution, report = routine.solve(A, b, ids, reuse, id=job.id)
                report.add(**aux)
                job.submit_solution(solution, report)
            return job
        
        def run_job_single(job: SimJob):
            for A, b, ids, reuse, aux in job.iter_Ab():
                solution, report = self.solveroutine.solve(A, b, ids, reuse, id=job.id)
                report.add(**aux)
                job.submit_solution(solution, report)
            return job
        
        ## GROUP FREQUENCIES
        # Each frequency group will be pre-assembled before submitting them to the parallel pool
        freq_groups = []
        if frequency_groups == -1:
            freq_groups=[self.frequencies,]
        else:
            n = frequency_groups
            freq_groups = [self.frequencies[i:i+n] for i in range(0, len(self.frequencies), n)]

        results: list[SimJob] = []
        matset: list[tuple[np.ndarray, np.ndarray, np.ndarray]] = []
        
        logger.trace(f'Frequency groups: {freq_groups}')
        ## Single threaded
        job_id = 1

        self._compute_modes(sum(self.frequencies)/len(self.frequencies))

        if not parallel:
            # ITERATE OVER FREQUENCIES
            freq_groups
            for i_group, fgroup in enumerate(freq_groups):
                logger.info(f'Precomputing group {i_group}.')
                jobs = []
                ## Assemble jobs
                for ifreq, freq in enumerate(fgroup):
                    logger.debug(f'Simulation frequency = {freq/1e9:.3f} GHz') 
                    if automatic_modal_analysis:
                        self._compute_modes(freq)
                    job, mats = self.assembler.assemble_freq_matrix(self.basis, materials, 
                                                            self.bc.boundary_conditions, 
                                                            freq, 
                                                            cache_matrices=self.cache_matrices)
                    job.store_limit = harddisc_threshold
                    job.relative_path = harddisc_path
                    job.id = job_id
                    job_id += 1
                    jobs.append(job)
                    matset.append(mats)
                
                logger.info(f'Starting single threaded solve of {len(jobs)} jobs.')
                group_results = [run_job_single(job) for job in jobs]
                results.extend(group_results)
        elif not multi_processing:
             # MULTI THREADED
            with ThreadPoolExecutor(max_workers=n_workers, thread_name_prefix='WKR') as executor:
                # ITERATE OVER FREQUENCIES
                for i_group, fgroup in enumerate(freq_groups):
                    logger.info(f'Precomputing group {i_group}.')
                    jobs = []
                    ## Assemble jobs
                    for freq in fgroup:
                        logger.debug(f'Simulation frequency = {freq/1e9:.3f} GHz') 
                        if automatic_modal_analysis:
                            self._compute_modes(freq)
                        job, mats = self.assembler.assemble_freq_matrix(self.basis, materials, 
                                                                self.bc.boundary_conditions, 
                                                                freq, 
                                                                cache_matrices=self.cache_matrices)
                        job.store_limit = harddisc_threshold
                        job.relative_path = harddisc_path
                        job.id = job_id
                        job_id += 1
                        jobs.append(job)
                        matset.append(mats)
                    
                    logger.info(f'Starting distributed solve of {len(jobs)} jobs with {n_workers} threads.')
                    group_results = list(executor.map(run_job, jobs))
                    results.extend(group_results)
                executor.shutdown()
        else:
            ### MULTI PROCESSING
            # Check for if __name__=="__main__" Guard
            if not called_from_main_function():
                raise SimulationError(
                    "Multiprocess support must be launched from your "
                    "if __name__ == '__main__' guard in the top-level script."
                )
            # Start parallel pool
            with mp.Pool(processes=n_workers, initializer=_init_worker) as pool:
                for i_group, fgroup in enumerate(freq_groups):
                    logger.debug(f'Precomputing group {i_group}.')
                    jobs = []
                    # Assemble jobs
                    for freq in fgroup:
                        logger.debug(f'Simulation frequency = {freq/1e9:.3f} GHz')
                        if automatic_modal_analysis:
                            self._compute_modes(freq)
                        
                        job, mats = self.assembler.assemble_freq_matrix(
                            self.basis, materials,
                            self.bc.boundary_conditions,
                            freq,
                            cache_matrices=self.cache_matrices
                        )

                        job.store_limit = harddisc_threshold
                        job.relative_path = harddisc_path
                        job.id = job_id
                        job_id += 1
                        jobs.append(job)
                        matset.append(mats)

                    logger.info(
                        f'Starting distributed solve of {len(jobs)} jobs '
                        f'with {n_workers} processes in parallel'
                    )
                    # Distribute taks
                    group_results = pool.map(run_job_multi, jobs)
                    results.extend(group_results)
        if parallel:
            thread_local.__dict__.clear()
        logger.info('Solving complete')

        for freq, job in zip(self.frequencies, results):
            self.data.setreport(job.reports, freq=freq, **self._params)

        for variables, data in self.data.sim.iterate():
            logger.trace(f'Sim variable: {variables}')
            for item in data['report']:
                item.logprint(logger.trace)

        self.solveroutine.reset()
        ### Compute S-parameters and return
        self._post_process(results, matset)
        return self.data
    
    def eigenmode(self, search_frequency: float,
                        nmodes: int = 6,
                        k0_limit: float = 1,
                        direct: bool = False,
                        deep_search: bool = False,
                        mode: Literal['LM','LR','SR','LI','SI']='LM') -> MWData:
        """Executes an eigenmode study

       

        Args:
            search_frequency (float): The frequency around which you would like to search
            nmodes (int, optional): The number of jobs. Defaults to 6.
            k0_limit (float): The lowest k0 value before which a mode is considered part of the null space. Defaults to 1e-3
        Raises:
            SimulationError: An error associated witha a problem during the simulation.

        Returns:
            MWSimData: The dataset.
        """
        
        self._simstart = time.time()
        if self.bc._initialized is False:
            raise SimulationError('Cannot run a modal analysis because no boundary conditions have been assigned.')
        
        self._initialize_field()
        self._initialize_bc_data()
        
        if self.basis is None:
            raise SimulationError('Cannot proceed. The simulation basis class is undefined.')

        materials = self.mesh._get_material_assignment(self.mesher.volumes)
        
        # er = self.mesh.retreive(lambda mat,x,y,z: mat.fer3d_mat(x,y,z), self.mesher.volumes)
        # ur = self.mesh.retreive(lambda mat,x,y,z: mat.fur3d_mat(x,y,z), self.mesher.volumes)
        # cond = self.mesh.retreive(lambda mat,x,y,z: mat.cond, self.mesher.volumes)[0,0,:]

        ### Does this move
        logger.debug('Initializing frequency domain sweep.')
            
        logger.info(f'Pre-assembling matrices of {len(self.frequencies)} frequency points.')
        
        job, matset = self.assembler.assemble_eig_matrix(self.basis, materials, 
                                                            self.bc.boundary_conditions, search_frequency)
        
        er, ur, cond = matset
        logger.info('Solving complete')

        A, C, solve_ids = job.yield_AC()

        target_k0 = 2*np.pi*search_frequency/299792458

        eigen_values, eigen_modes, report = self.solveroutine.eig(A, C, solve_ids, nmodes, direct, target_k0, which=mode)

        eigen_modes = job.fix_solutions(eigen_modes)

        logger.debug(f'Eigenvalues: {np.sqrt(eigen_values)} rad/m')

        nmodes_found = eigen_values.shape[0]

        for i in range(nmodes_found):
            

            eig_k0 = np.sqrt(eigen_values[i])
            if eig_k0 < k0_limit:
                logger.debug(f'Ignoring mode due to low k0: {eig_k0} < {k0_limit}')
                continue
            eig_freq = eig_k0*299792458/(2*np.pi)

            logger.debug(f'Found k0={eig_k0:.2f}, f0={eig_freq/1e9:.2f} GHz')
            Emode = eigen_modes[:,i]

            scalardata = self.data.scalar.new(**self._params)
            scalardata.k0 = eig_k0
            scalardata.freq = eig_freq

            fielddata = self.data.field.new(**self._params)
            fielddata.freq = eig_freq
            fielddata._der = np.squeeze(er[0,0,:])
            fielddata._dur = np.squeeze(ur[0,0,:])
            fielddata._mode_field = Emode
            fielddata.basis = self.basis
        ### Compute S-parameters and return
        
        return self.data

    def _post_process(self, results: list[SimJob], materials: list[tuple[np.ndarray, np.ndarray, np.ndarray]]):
        """Compute the S-parameters after Frequency sweep

        Args:
            results (list[SimJob]): The set of simulation results
            er (np.ndarray): The domain εᵣ
            ur (np.ndarray): The domain μᵣ
            cond (np.ndarray): The domain conductivity
        """
        if self.basis is None:
            raise SimulationError('Cannot post-process. Simulation basis function is undefined.')
        mesh = self.mesh
        all_ports = self.bc.oftype(PortBC)
        port_numbers = [port.port_number for port in all_ports]
        
        logger.info('Computing S-parameters')
        

        for freq, job, mats in zip(self.frequencies, results, materials):
            er, ur, cond = mats
            ertri = np.zeros((3,3,self.mesh.n_tris), dtype=np.complex128)
            urtri = np.zeros((3,3,self.mesh.n_tris), dtype=np.complex128)
            condtri = np.zeros((self.mesh.n_tris,), dtype=np.complex128)

            for itri in range(self.mesh.n_tris):
                itet = self.mesh.tri_to_tet[0,itri]
                ertri[:,:,itri] = er[:,:,itet]
                urtri[:,:,itri] = ur[:,:,itet]
                condtri[itri] = cond[0,0,itet]
                
            k0 = 2*np.pi*freq/299792458

            scalardata = self.data.scalar.new(freq=freq, **self._params)
            scalardata.k0 = k0
            scalardata.freq = freq
            scalardata.init_sp(port_numbers) # type: ignore
            
            fielddata = self.data.field.new(freq=freq, **self._params)
            fielddata.freq = freq
            fielddata._der = np.squeeze(er[0,0,:])
            fielddata._dur = np.squeeze(ur[0,0,:])

            logger.info(f'Post Processing simulation frequency = {freq/1e9:.3f} GHz') 

            # Recording port information
            for active_port in all_ports:
                port_tets = self.mesh.get_face_tets(active_port.tags)
                
                fielddata.add_port_properties(active_port.port_number,
                                         mode_number=active_port.mode_number,
                                         k0 = k0,
                                         beta = active_port.get_beta(k0),
                                         Z0 = active_port.portZ0(k0),
                                         Pout = active_port.power)
                scalardata.add_port_properties(active_port.port_number,
                                         mode_number=active_port.mode_number,
                                         k0 = k0,
                                         beta = active_port.get_beta(k0),
                                         Z0 = active_port.portZ0(k0),
                                         Pout= active_port.power)

                # Set port as active and add the port mode to the forcing vector
                active_port.active = True
                
                solution = job._fields[active_port.port_number]

                fielddata._fields = job._fields
                fielddata.basis = self.basis
                # Compute the S-parameters
                # Define the field interpolation function
                fieldf = self.basis.interpolate_Ef(solution, tetids=port_tets)
                Pout = 0.0 + 0j

                # Active port power
                tris = mesh.get_triangles(active_port.tags)
                tri_vertices = mesh.tris[:,tris]
                pfield, pmode = self._compute_s_data(active_port, fieldf, tri_vertices, k0, ertri[:,:,tris], urtri[:,:,tris])
                logger.debug(f'[{active_port.port_number}] Active port amplitude = {np.abs(pfield):.3f} (Excitation = {np.abs(pmode):.2f})')
                Pout = pmode
                
                #Passive ports
                for bc in all_ports:
                    port_tets = self.mesh.get_face_tets(bc.tags)
                    fieldf = self.basis.interpolate_Ef(solution, tetids=port_tets)
                    tris = mesh.get_triangles(bc.tags)
                    tri_vertices = mesh.tris[:,tris]
                    pfield, pmode = self._compute_s_data(bc, fieldf,tri_vertices, k0, ertri[:,:,tris], urtri[:,:,tris])
                    logger.debug(f'[{bc.port_number}] Passive amplitude = {np.abs(pfield):.3f}')
                    scalardata.write_S(bc.port_number, active_port.port_number, pfield/Pout)
                active_port.active=False
            
            fielddata.set_field_vector()

        logger.info('Simulation Complete!')
        self._simend = time.time()    
        logger.info(f'Elapsed time = {(self._simend-self._simstart):.2f} seconds.')

    
    def _compute_s_data(self, bc: PortBC, 
                       fieldfunction: Callable, 
                       tri_vertices: np.ndarray, 
                       k0: float,
                       erp: np.ndarray,
                       urp: np.ndarray,) -> tuple[complex, complex]:
        """ Computes the S-parameter data for a given boundary condition and field function.

        Args:
            bc (PortBC): The port boundary condition
            fieldfunction (Callable): The field function that interpolates the solution field.
            tri_vertices (np.ndarray): The triangle vertex indices of the port face
            k₀ (float): The simulation phase constant
            erp (np.ndarray): The εᵣ of the port face triangles
            urp (np.ndarray): The μᵣ of the port face triangles.

        Returns:
            tuple[complex, complex]: _description_
        """
        from .sparam import sparam_field_power, sparam_mode_power
        if bc.v_integration:
            if bc.vintline is None:
                raise SimulationError('Trying to compute characteristic impedance but no integration line is defined.')
            if bc.Z0 is None:
                raise SimulationError('Trying to compute the impedance of a boundary condition with no characteristic impedance.')
            
            Voltages = [line.line_integral(fieldfunction) for line in bc.vintline]
            V = sum(Voltages)/len(Voltages)
            
            if bc.active:
                if bc.voltage is None:
                    raise ValueError('Cannot compute port S-paramer with a None port voltage.')
                a = bc.voltage
                b = (V-bc.voltage)
            else:
                a = 0
                b = V
            
            a_sig = a*csqrt(1/(2*bc.Z0))
            b_sig = b*csqrt(1/(2*bc.Z0))

            return b_sig, a_sig
        else:
            if bc.modetype(k0) == 'TEM':
                const = 1/(np.sqrt((urp[0,0,:] + urp[1,1,:] + urp[2,2,:])/(erp[0,0,:] + erp[1,1,:] + erp[2,2,:])))
            if bc.modetype(k0) == 'TE':
                const = 1/((urp[0,0,:] + urp[1,1,:] + urp[2,2,:])/3)
            elif bc.modetype(k0) == 'TM':
                const = 1/((erp[0,0,:] + erp[1,1,:] + erp[2,2,:])/3)
            const = np.squeeze(const)
            field_p = sparam_field_power(self.mesh.nodes, tri_vertices, bc, k0, fieldfunction, const, 5)
            mode_p = sparam_mode_power(self.mesh.nodes, tri_vertices, bc, k0, const, 5)
            return field_p, mode_p


    ############################################################
    #                     DEPRICATED FUNCTIONS                #
    ############################################################

    def frequency_domain(self, *args, **kwargs):
        """DEPRICATED VERSION: Use run_sweep() instead.
        """
        logger.warning('This function is depricated. Please use run_sweep() instead')
        return self.run_sweep(*args, **kwargs)


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import gmsh # type: ignore
from .geometry import GeoVolume, GeoObject, GeoSurface
from .selection import Selection, FaceSelection
from .periodic import PeriodicCell
import numpy as np
from typing import Iterable, Callable, Any, TypeVar
from loguru import logger
from enum import Enum
from .bc import Periodic

class MeshError(Exception):
    pass

class Algorithm2D(Enum):
    MESHADAPT = 1
    AUTOMATIC = 2
    INITIAL_MESH_ONLY = 3
    DELAUNAY = 5
    FRONTAL_DELAUNAY = 6
    BAMG = 7
    FRONTAL_DELAUNAY_QUADS = 8
    PACKING_PARALLELOGRAMS = 9
    QUASI_STRUCTURED_QUAD = 11

#(1: Delaunay, 3: Initial mesh only, 4: Frontal, 7: MMG3D, 9: R-tree, 10: HXT)

class Algorithm3D(Enum):
    DELAUNAY = 1
    INITIAL_MESH_ONLY = 3
    FRONTAL = 4
    MMG3D = 7
    RTREE = 9
    HXT = 10

_DOM_TO_STR = {
    0: "point",
    1: "edge",
    2: "face",
    3: "volume",
}

T = TypeVar('T')
def unpack_lists(_list: Any, collector: list | None = None) -> list[Any]:
    '''Unpack a recursive list of lists'''
    if collector is None:
        collector = []
    for item in _list:
        if isinstance(item, list):
            unpack_lists(item, collector)
        else:
            collector.append(item)
    
    return collector

class Mesher:

    def __init__(self):
        self.objects: list[GeoObject] = []
        self.size_definitions: list[tuple[int, float]] = []
        self.mesh_fields: list[int] = []
        self.min_size: float = None
        self.max_size: float = None
        self.periodic_cell: PeriodicCell = None

        
    @property
    def edge_tags(self) -> list[int]:
        return [tag[1] for tag in gmsh.model.getEntities(1)]
    
    @property
    def face_tags(self) -> list[int]:
        return [tag[1] for tag in gmsh.model.getEntities(2)]
    
    @property
    def node_tags(self) -> list[int]:
        return [tag[1] for tag in gmsh.model.getEntities(0)]
    
    @property
    def volumes(self) -> list[GeoVolume]:
        return [obj for obj in self.objects if isinstance(obj, GeoVolume)]
    
    @property
    def domain_boundary_face_tags(self) -> list[int]:
        '''Get the face tags of the domain boundaries'''
        domain_tags = gmsh.model.getEntities(3)
        tags = gmsh.model.getBoundary(domain_tags, combined=True, oriented=False)
        return [int(tag[1]) for tag in tags]
    
    @property
    def domain_internal_face_tags(self) -> list[int]:
        alltags = self.face_tags
        boundary = self.domain_boundary_face_tags
        return [tag for tag in alltags if tag not in boundary]
    
    def _get_periodic_bcs(self) -> list[Periodic]:
        if self.periodic_cell is None:
            return []
        return self.periodic_cell._bcs
    
    def _check_ready(self) -> None:
        if self.max_size is None or self.min_size is None:
            raise MeshError('Either maximum or minimum mesh size is undefined. Make sure \
                            to set the simulation frequency range before calling mesh instructions.')
    
    def submit_objects(self, objects: GeoObject | list[GeoObject] | list[list[GeoObject]]) -> None:
        """Takes al ist of GeoObjects and computes the fragment. 

        Args:
            objects (list[GeoObject]): The set of GeoObjects
        """
        if not isinstance(objects, list):
            objects = [objects,]

        objects = unpack_lists(objects)
        embeddings: list = []
        gmsh.model.occ.synchronize()

        final_dimtags = unpack_lists([domain.dimtags for domain in objects]) # type: ignore

        dom_mapping = dict()
        for dom in objects: # type: ignore
            embeddings.extend(dom._embeddings) # type: ignore
            for dt in dom.dimtags: # type: ignore
                dom_mapping[dt] = dom
        

        embedding_dimtags = unpack_lists([emb.dimtags for emb in embeddings])

        tag_mapping: dict[int, dict] = {0: dict(),
                                        1: dict(),
                                        2: dict(),
                                        3: dict()}
        if len(objects) > 0: # type: ignore
            dimtags, output_mapping = gmsh.model.occ.fragment(final_dimtags, embedding_dimtags)

            for domain, mapping in zip(final_dimtags + embedding_dimtags, output_mapping):
                tag_mapping[domain[0]][domain[1]] = [o[1] for o in mapping]
            for dom in objects: # type: ignore
                dom.update_tags(tag_mapping) # type: ignore
        else:
            dimtags = final_dimtags
        
        self.objects = objects # type: ignore
        
        gmsh.model.occ.synchronize()

    def _set_mesh_periodicity(self, 
                     face1: Selection,
                     face2: Selection,
                     lattice: np.ndarray):
        translation = [1,0,0,lattice[0],
                       0,1,0,lattice[1],
                       0,0,1,lattice[2],
                       0,0,0,1]
        gmsh.model.mesh.set_periodic(2, face2.tags, face1.tags, translation)

    def set_algorithm(self,
                      obj: GeoObject,
                      algorithm: Algorithm3D) -> None:
        for tag in obj.tags:
            gmsh.model.mesh.setAlgorithm(obj.dim, tag, algorithm.value)
            
    def set_periodic_cell(self, cell: PeriodicCell, excluded_faces: Selection | None = None):
        """Sets the periodic cell information based on the PeriodicCell class object"""
        if excluded_faces is None:
            for f1, f2, lat in cell.cell_data():
                self._set_mesh_periodicity(f1, f2, lat)
        else:
            for f1, f2, lat in cell.cell_data():
                self._set_mesh_periodicity(f1 - excluded_faces, f2 - excluded_faces, lat)
        self.periodic_cell = cell

    def _set_size_in_domain(self, tags: list[int], max_size: float) -> None:
        """Define the size of the mesh inside a domain

        Args:
            tags (list[int]): The tags of the geometry
            max_size (float): The maximum size (in meters)
        """
        ctag = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.set_numbers(ctag, "VolumesList", tags)
        gmsh.model.mesh.field.set_number(ctag, "VIn", max_size)
        self.mesh_fields.append(ctag)

    def _set_size_on_face(self, tags: list[int], max_size: float) -> None:
        """Define the size of the mesh on a face

        Args:
            tags (list[int]): The tags of the geometry
            max_size (float): The maximum size (in meters)
        """
        ctag = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.set_numbers(ctag, "SurfacesList", tags)
        gmsh.model.mesh.field.set_number(ctag, "VIn", max_size)
        self.mesh_fields.append(ctag)
        
    def _set_size_on_edge(self, tags: list[int], max_size: float) -> None:
        """Define the size of the mesh on an edge

        Args:
            tags (list[int]): The tags of the geometry
            max_size (float): The maximum size (in meters)
        """
        ctag = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.set_numbers(ctag, "CurvesList", tags)
        gmsh.model.mesh.field.set_number(ctag, "VIn", max_size)
        self.mesh_fields.append(ctag)
        
    def _set_size_on_point(self, tags: list[int], max_size: float) -> None:
        """Define the size of the mesh on a point

        Args:
            tags (list[int]): The tags of the geometry
            max_size (float): The maximum size (in meters)
        """
        ctag = gmsh.model.mesh.field.add("Constant")
        gmsh.model.mesh.field.set_numbers(ctag, "PointsList", tags)
        gmsh.model.mesh.field.set_number(ctag, "VIn", max_size)
        self.mesh_fields.append(ctag)

    def _configure_mesh_size(self, discretizer: Callable, resolution: float):
        """Defines the mesh sizes based on a discretization callable.
        The discretizer must take a material and return a maximum
        size for that material.

        Args:
            discretizer (Callable): The discretization function
            resolution (float): The resolution
        """
        logger.debug('Starting initial mesh size computation.')
        dimtags = gmsh.model.occ.get_entities(2)

        for dim, tag in dimtags:
            gmsh.model.mesh.setSizeFromBoundary(2, tag, 0)

        mintag = gmsh.model.mesh.field.add("Min")
        size_mapping = dict()

        for obj in sorted(self.objects, key=lambda x: x._priority):
            if obj._unset_constraints:
                self.unset_constraints(obj.dimtags)

            size = discretizer(obj.material)*resolution*obj.mesh_multiplier
            size = min(size, obj.max_meshsize)
            
            for tag in obj.tags:
                size_mapping[tag] = size
        
        for tag, size in size_mapping.items():
            logger.debug(f'Setting mesh size:{1000*size:.3f}mm in domains: {tag}')
            self._set_size_in_domain([tag,], size)

        gmsh.model.mesh.field.setNumbers(mintag, "FieldsList", self.mesh_fields)
        gmsh.model.mesh.field.setAsBackgroundMesh(mintag)

        for tag, size in self.size_definitions:
            logger.debug(f'Setting aux size definition: {1000*size:.3f}mm in domain {tag}.')
            gmsh.model.mesh.setSize([tag,], size)

    def unset_constraints(self, dimtags: list[tuple[int,int]]):
        '''Unset the mesh constraints for the given dimension tags.'''
        logger.trace(f'Unsetting mesh size constraint for domains: {dimtags}')
        for dimtag in dimtags:
            gmsh.model.mesh.setSizeFromBoundary(dimtag[0], dimtag[1], 0)
            
    def set_boundary_size(self, 
                          boundary: GeoObject | Selection | Iterable, 
                          size:float,
                          growth_rate: float = 1.4,
                          max_size: float | None = None) -> None:
        """Refine the mesh size along the boundary of a conducting surface

        Args:
            object (GeoSurface | FaceSelection): _description_
            size (float): _description_
            growth_rate (float, optional): _description_. Defaults to 1.1.
            max_size (float, optional): _description_. Defaults to None.
        """
        if isinstance(boundary, Iterable):
            for bound in boundary:
                self.set_boundary_size(bound, size, growth_rate, max_size)
            return
        
        dimtags = boundary.dimtags
        
        
        if max_size is None:
            self._check_ready()
            max_size = self.max_size
        
        growth_distance = np.log10(max_size/size)/np.log10(growth_rate)
        
        logger.debug(f'Setting boundary size for region {dimtags} to {size*1000:.3f}mm, GR={growth_rate:.3f}, dist={growth_distance*1000:.2f}mm, Max={max_size*1000:.3f}mm')
        
        nodes = gmsh.model.getBoundary(dimtags, combined=False, oriented=False, recursive=False)

        disttag = gmsh.model.mesh.field.add("Distance")
        if boundary.dim==2:
            gmsh.model.mesh.field.setNumbers(disttag, "CurvesList", [n[1] for n in nodes])
        if boundary.dim==3:
            gmsh.model.mesh.field.setNumbers(disttag,'SurfacesList', [n[1] for n in nodes])
        gmsh.model.mesh.field.setNumber(disttag, "Sampling", 100)

        thtag = gmsh.model.mesh.field.add("Threshold")
        gmsh.model.mesh.field.setNumber(thtag, "InField", disttag)
        gmsh.model.mesh.field.setNumber(thtag, "SizeMin", size)
        gmsh.model.mesh.field.setNumber(thtag, "SizeMax", max_size)
        gmsh.model.mesh.field.setNumber(thtag, "DistMin", size)
        gmsh.model.mesh.field.setNumber(thtag, "DistMax", growth_distance*size)
    
        self.mesh_fields.append(thtag)

    def set_domain_size(self, obj: GeoObject | Selection, size: float):
        """Manually set the maximum element size inside a domain

        Args:
            obj (GeoVolume | Selection): The volumetric domain
            size (float): The maximum mesh size
        """
        if obj.dim != 3:
            logger.warning('Provided object is not a volume.')
            if obj.dim==2:
                logger.warning('Forwarding to set_face_size')
                self.set_face_size(obj, size)
        logger.debug(f'Setting size {size*1000:.3f}ff for object {obj}')
        self._set_size_in_domain(obj.tags, size)

    def set_face_size(self, obj: GeoSurface | Selection, size: float):
        """Manually set the maximum element size on a face

        Args:
            obj (GeoSurface | Selection): The surface domain
            size (float): The maximum size
        """
        if obj.dim != 2:
            logger.warning('Provided object is not a surface.')
            if obj.dim==3:
                logger.warning('Forwarding to set_domain_size')
                self.set_face_size(obj, size)
        
        logger.debug(f'Setting size {size*1000:.3f}ff for face {obj}')
        self._set_size_on_face(obj.tags, size)
    
    def set_size(self, obj: GeoObject, size: float) -> None:
        """Manually set the size in or on an object

        Args:
            obj (GeoObject): _description_
            size (float): _description_
        """
        if obj.dim == 2:
            self._set_size_on_face(obj.tags, size)
        elif obj.dim == 3:
            self._set_size_in_domain(obj.tags, size)
        elif obj.dim == 1:
            self._set_size_on_edge(obj.tags, size)
        elif obj.dim == 0:
            self._set_size_on_point(obj.tags, size)
        
    def refine_conductor_edge(self, dimtags: list[tuple[int,int]], size):
        nodes = gmsh.model.getBoundary(dimtags, combined=False, recursive=False)

        # for node in nodes:
        #     pcoords = np.linspace(0, 0.5, 10)
        #     gmsh.model.mesh.setSizeAtParametricPoints(node[0], node[1], pcoords, size*np.ones_like(pcoords))
        #     #self.size_definitions.append((node, size))
        # gmsh.model.mesh.setSizeFromBoundary(dimtag[0], dimtag[1], 0)

        tag = gmsh.model.mesh.field.add("Distance")

        #gmsh.model.mesh.field.setNumbers(1, "PointsList", [5])
        gmsh.model.mesh.field.setNumbers(tag, "CurvesList", [n[1] for n in nodes])
        gmsh.model.mesh.field.setNumber(tag, "Sampling", 100)

        # We then define a `Threshold' field, which uses the return value of the
        # `Distance' field 1 in order to define a simple change in element size
        # depending on the computed distances
        #
        # SizeMax -                     /------------------
        #                              /
        #                             /
        #                            /
        # SizeMin -o----------------/
        #          |                |    |
        #        Point         DistMin  DistMax
        thtag = gmsh.model.mesh.field.add("Threshold")
        gmsh.model.mesh.field.setNumber(thtag, "InField", tag)
        gmsh.model.mesh.field.setNumber(thtag, "SizeMin", size)
        gmsh.model.mesh.field.setNumber(thtag, "SizeMax", 100)
        gmsh.model.mesh.field.setNumber(thtag, "DistMin", 0.2*size)
        gmsh.model.mesh.field.setNumber(thtag, "DistMax", 5*size)

        self.mesh_fields.append(thtag)
        

        for dimtag in dimtags:
            gmsh.model.mesh.setSizeFromBoundary(dimtag[0], dimtag[1], 0)




# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import warnings
from loguru import logger

# Catch the Cuda warning and print it with Loguru.
with warnings.catch_warnings(record=True) as w:
    warnings.simplefilter("always")
    import cupy as cp
    for warn in w:
        logger.debug(f"{warn.category.__name__}: {warn.message}")

import nvmath.bindings.cudss as cudss # ty: ignore
from nvmath import CudaDataType # ty: ignore

from scipy.sparse import csr_matrix
import numpy as np
from typing import Literal

############################################################
#                         CONSTANTS                        #
############################################################

ALG_NEST_DISS_METIS = cudss.AlgType.ALG_DEFAULT
ALG_COLAMD = cudss.AlgType.ALG_1
ALG_COLAMD_BLOCK_TRI = cudss.AlgType.ALG_2
ALG_AMD = cudss.AlgType.ALG_3

FLOAT64 = CudaDataType.CUDA_R_64F
FLOAT32 = CudaDataType.CUDA_R_32F
COMPLEX128 = CudaDataType.CUDA_C_64F
COMPLEX64 = CudaDataType.CUDA_C_32F
INT64 = CudaDataType.CUDA_R_64I
INT32 = CudaDataType.CUDA_R_32I

INDEX_BASE = cudss.IndexBase.ZERO

def _c_pointer(arry) -> int:
    return int(arry.data.ptr)

############################################################
#                         INTERFACE                        #
############################################################

class CuDSSInterface:
    """the CuDSSInterface class implements the nvmath bindings and cupy
    control for EMerge.
    """
    AlgType = cudss.AlgType

    def __init__(self):
        self.A_cu = None
        self.b_cu = None
        self.x_cu = None
        self.A_cobj = None
        self.b_cobj = None
        self.x_cobj = None
        self.A_pattern = None

        self._handle = cudss.create()
        self._config = cudss.config_create()
        self._data = cudss.data_create(self._handle)

        self.MTYPE = cudss.MatrixType.SYMMETRIC
        self.MVIEW = cudss.MatrixViewType.FULL
        self.RALG = cudss.AlgType.ALG_DEFAULT
        self.VTYPE = CudaDataType.CUDA_R_64F

        self._INDPTR = None
        self._ROW_START: int | None = None
        self._ROW_END: int | None = None
        self._IND = None
        self._VAL = None
        self._NNZ: int | None = None
        self._COMP: bool = True
        self._PRES: int = 2
        self._COL_IDS = None

        self._initialized = False

        param = cudss.ConfigParam.REORDERING_ALG
        dtype = cudss.get_config_param_dtype(int(param))
        reorder_alg = np.array(self.RALG, dtype=dtype)

        cudss.config_set(
            self._config,   
            int(param),
            reorder_alg.ctypes.data, 
            reorder_alg.nbytes
        )

    def set_algorithm(self, alg_type: Literal['METIS','COLAMD','COLAM_BT','AMD']):
        """Define fill-in reduction column permuation algorithm. The options are:

         - "METIS" (Default) = NVidia's own Nested Dissection METIS sorter
         - "COLAMD" = Column approximate minimum degree
         - "COLAM_BT" = Column Approximate Minimum Degree Block Triangular
         - "AMD" = Approximate Minimum Degree

        Args:
            alg_type (str): The chosen type
        """
        if alg_type=='METIS':
            self.RALG = ALG_NEST_DISS_METIS
        elif alg_type =='COLAMD':
            self.RALG = ALG_COLAMD
        elif alg_type == 'COLAMD_BT':
            self.RALG = ALG_COLAMD_BLOCK_TRI
        elif alg_type == 'AMD':
            self.RALG = ALG_AMD
        else:
            logger.warning(f'Algorithm type {alg_type} is not of the chosen set. Ignoring setting.')
    
    def init_type(self):
        """Initializes the value data type of the solver (float vs complex, single vs double).
        """
        if self._PRES == 1:
            if self._COMP:
                self.c_dtype = cp.complex64
                self.VTYPE = COMPLEX64
            else:
                self.c_dtype = cp.float32
                self.VTYPE = FLOAT32
        else:
            if self._COMP:
                self.c_dtype = cp.complex128
                self.VTYPE = COMPLEX128
            else:
                self.c_dtype = cp.float64
                self.VTYPE = FLOAT64

    def submit_matrix(self, A: csr_matrix):
        """Sets the given csr_matrix as the matrix to be solved.

        Args:
            A (csr_matrix): The csr_format matrix for the problem Ax=b
        """
        self.N = A.shape[0]

        if np.iscomplexobj(A):
            self._COMP = True
        else:
            self._COMP = False
        
        self.init_type()

        self.A_cu = cp.sparse.csr_matrix(A).astype(self.c_dtype)

        self._INDPTR = cp.ascontiguousarray(self.A_cu.indptr.astype(cp.int32))
        self._IND = cp.ascontiguousarray(self.A_cu.indices.astype(cp.int32))
        self._VAL = cp.ascontiguousarray(self.A_cu.data)
        self._NNZ = int(self._VAL.size)
        self._ROW_START = self._INDPTR[:-1]
        self._ROW_END = self._INDPTR[1:]
        self._COL_IDS = self.A_cu.indices.astype(cp.int32)

    def submit_vector(self, b: np.ndarray):
        """Submits the dense vector b to be solved.

        Args:
            b (np.ndarray): The dense vector for the problem Ax=b
        """
        self.b_cu = cp.array(b).astype(self.c_dtype)
    
    def create_solvec(self):
        """Initializes a solution vector that the nvmath binding can access.
        """
        self.x_cu = cp.empty_like(self.b_cu)

    def _update_dss_data(self):
        """Updates the currently defined matrix data into the existing memory.ALG_AMD
        """
        cudss.matrix_set_values(self.A_cobj, _c_pointer(self._VAL))

        
        self.b_cobj = cudss.matrix_create_dn(self.N, 1, self.N, _c_pointer(self.b_cu),
                                    int(self.VTYPE), int(cudss.Layout.COL_MAJOR))
        self.x_cobj = cudss.matrix_create_dn(self.N, 1, self.N, _c_pointer(self.x_cu),
                                    int(self.VTYPE), int(cudss.Layout.COL_MAJOR))

    def _create_dss_data(self):
        """Creates a new memory slot for the CSR matrix of the matrix A"""
        self.A_cobj = cudss.matrix_create_csr(
            self.N,self.N,self._NNZ,
            _c_pointer(self._ROW_START),
            _c_pointer(self._ROW_END),
            _c_pointer(self._COL_IDS),
            _c_pointer(self._VAL),
            int(INT32), 
            int(self.VTYPE),
            int(self.MTYPE),
            int(self.MVIEW),
            int(INDEX_BASE),
        )

        self.b_cobj = cudss.matrix_create_dn(self.N, 1, self.N, _c_pointer(self.b_cu),
                                    int(self.VTYPE), int(cudss.Layout.COL_MAJOR))
        self.x_cobj = cudss.matrix_create_dn(self.N, 1, self.N, _c_pointer(self.x_cu),
                                    int(self.VTYPE), int(cudss.Layout.COL_MAJOR))

    def from_symbolic(self, A: csr_matrix, b: np.ndarray) -> np.ndarray:
        """Solves Ax=b starting from the symbolic factorization

        Args:
            A (csr_matrix): The input sparse matrix
            b (np.ndarray): The solution vector b

        Returns:
            np.ndarray: The solved vector
        """
        self.submit_matrix(A)
        self.submit_vector(b)
        self.create_solvec()
        self._create_dss_data()
        self._symbolic()
        self._numeric(False)
        return self._solve()

    def from_numeric(self, A: csr_matrix, b: np.ndarray) -> np.ndarray:
        """Solves Ax=b starting from the Numeric factorization

        Args:
            A (csr_matrix): The input sparse matrix
            b (np.ndarray): The solution vector b

        Returns:
            np.ndarray: The solved vector
        """
        self.submit_matrix(A)
        self.submit_vector(b)
        self.create_solvec()
        self._update_dss_data()
        self._numeric(True)
        return self._solve()

    def from_solve(self, b: np.ndarray) -> np.ndarray:
        """Solves Ax=b only with a new b vector.

        Args:
            A (csr_matrix): The input sparse matrix
            b (np.ndarray): The solution vector b

        Returns:
            np.ndarray: The solved vector
        """
        self.submit_vector(b)
        self.create_solvec()
        return self._solve()

    def _symbolic(self):
        logger.trace('Executing symbolic factorization')
        cudss.execute(self._handle, cudss.Phase.ANALYSIS, self._config, self._data, 
                      self.A_cobj, self.x_cobj, self.b_cobj)

    def _numeric(self, refactorize: bool = False):
        if refactorize:
            logger.trace('Refactoring matrix')
            phase = cudss.Phase.REFACTORIZATION
        else:
            phase = cudss.Phase.FACTORIZATION
        logger.trace('Executing numerical factorization')
        cudss.execute(self._handle, phase, self._config, self._data, 
                      self.A_cobj, self.x_cobj, self.b_cobj)
                      
    def _solve(self) -> np.ndarray:
        logger.trace('Solving matrix problem')
        cudss.execute(self._handle, cudss.Phase.SOLVE, self._config, self._data, 
                      self.A_cobj, self.x_cobj, self.b_cobj)
        cp.cuda.runtime.deviceSynchronize()
        x_host = cp.asnumpy(self.x_cu).ravel()
        return x_host



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
from .mesher import Mesher, unpack_lists
from .geometry import GeoObject, _GEOMANAGER
from .geo.modeler import Modeler
from .physics.microwave.microwave_3d import Microwave3D
from .mesh3d import Mesh3D
from .selection import Selector, FaceSelection, Selection
from .logsettings import LOG_CONTROLLER
from .plot.pyvista import PVDisplay
from .dataset import SimulationDataset
from .periodic import PeriodicCell
from .cacherun import get_build_section, get_run_section
from .settings import DEFAULT_SETTINGS, Settings
from .solver import EMSolver, Solver
from typing import Literal, Generator, Any
from loguru import logger
import numpy as np
import gmsh # type: ignore
import os
import inspect
from pathlib import Path
import joblib
from atexit import register
import signal
from .. import __version__

############################################################
#                   EXCEPTION DEFINITIONS                  #
############################################################

_GMSH_ERROR_TEXT = """
--------------------------
Known problems/solutions:
(1) - PLC Error:  A segment and a facet intersect at point
    This can be caused when approximating thin curved volumes. Try to decrease the mesh size for that region.
--------------------------
"""

class SimulationError(Exception):
    pass

class VersionError(Exception):
    pass

############################################################
#                 BASE 3D SIMULATION MODEL                 #
############################################################

class Simulation:

    def __init__(self, 
                 modelname: str, 
                 loglevel: Literal['TRACE','DEBUG','INFO','WARNING','ERROR'] = 'INFO',
                 load_file: bool = False,
                 save_file: bool = False,
                 write_log: bool = False,
                 path_suffix: str = ".EMResults"):
        """Generate a Simulation class object.

        As a minimum a file name should be provided. Additionally you may provide it with any
        class that inherits from BaseDisplay. This will then be used for geometry displaying.

        Args:
            modelname (str): The name of the simulation model. This will be used for filenames and path names when saving data.
            loglevel ("DEBUG","INFO","WARNING","ERROR", optional): The loglevel to use for loguru. Defaults to 'INFO'.
            load_file (bool, optional): If the simulatio model should be loaded from a file. Defaults to False.
            save_file (bool, optional): if the simulation file should be stored to a file. Defaults to False.
            write_log (bool, optional): If a file should be created that contains the entire log of the simulation. Defaults to False.
            path_suffix (str, optional): The suffix that will be added to the results directory. Defaults to ".EMResults".
        """

        caller_file = Path(inspect.stack()[1].filename).resolve()
        base_path = caller_file.parent

        self.modelname = modelname
        self.modelpath = base_path / (modelname.lower()+path_suffix)
        self.mesher: Mesher = Mesher()
        self.modeler: Modeler = Modeler()
        
        self.mesh: Mesh3D = Mesh3D(self.mesher)
        self.select: Selector = Selector()
        
        self.settings: Settings = DEFAULT_SETTINGS

        ## Display
        self.display: PVDisplay = PVDisplay(self.mesh)
        
        ## Dataset
        self.data: SimulationDataset = SimulationDataset()
        
        ## STATES
        self.__active: bool = False
        self._defined_geometries: bool = False
        self._cell: PeriodicCell | None = None
        self.save_file: bool = save_file
        self.load_file: bool = load_file
        self._cache_run: bool = False
        self._file_lines: str = ''
        
        ## Physics
        self.mw: Microwave3D = Microwave3D(self.mesher, self.settings, self.data.mw)

        self._initialize_simulation()

        self.set_loglevel(loglevel)
        
        if write_log:
            self.set_write_log()

        LOG_CONTROLLER._flush_log_buffer()
        
        LOG_CONTROLLER._sys_info()
        
        self._update_data()
    

    ############################################################
    #                       PRIVATE FUNCTIONS                  #
    ############################################################

    def __setitem__(self, name: str, value: Any) -> None:
        """Store data in the current data container"""
        self.data.sim[name] = value

    def __getitem__(self, name: str) -> Any:
        """Get the data from the current data container"""
        return self.data.sim[name]
    
    def __enter__(self) -> Simulation:
        """This method is depricated with the new atexit system. It still exists for backwards compatibility.

        Returns:
            Simulation: the Simulation object
        """
        return self

    def __exit__(self, type, value, tb):
        """This method no longer does something. It only serves as backwards compatibility."""
        self._exit_gmsh()
        return False
    
    def _install_signal_handlers(self):
        # on SIGINT (Ctrl-C) or SIGTERM, call our exit routine
        for sig in (signal.SIGINT, signal.SIGTERM):
            signal.signal(sig, self._handle_signal)
    
    def _handle_signal(self, signum, frame):
        """
        Signal handler: do our cleanup, then re-raise
        the default handler so that exit code / traceback
        is as expected.
        """
        try:
            # run your atexit-style cleanup
            self._exit_gmsh()
        except Exception:
            # log but don’t block shutdown
            logger.exception("Error during signal cleanup")
        finally:
            # restore default handler and re‐send the signal
            signal.signal(signum, signal.SIG_DFL)
            os.kill(os.getpid(), signum)

    def _initialize_simulation(self):
        """Initializes the Simulation data and GMSH API with proper shutdown routines.
        """
        _GEOMANAGER.sign_in(self.modelname)
        
        # If GMSH is not yet initialized (Two simulation in a file)
        if gmsh.isInitialized() == 0:
            logger.debug('Initializing GMSH')
            gmsh.initialize()
            
            # Set an exit handler for Ctrl+C cases
            self._install_signal_handlers()

            # Restier the Exit GMSH function on proper program abortion
            register(self._exit_gmsh)
        else:
            gmsh.finalize()
            gmsh.initialize()
            
        # Create a new GMSH model or load it
        if not self.load_file:
            gmsh.model.add(self.modelname)
            self.data: SimulationDataset = SimulationDataset()
        else:
            self.load()

        # Set the Simulation state to active
        self.__active = True
        return self

    def _exit_gmsh(self):
        # If the simulation object state is still active (GMSH is running)
        if not self.__active:
            return
        logger.debug('Exiting program')
        # Save the file first
        if self.save_file:
            self.save()
        # Finalize GMSH
        if gmsh.isInitialized():
            gmsh.finalize()

        logger.debug('GMSH Shut down successful')
        # set the state to active
        self.__active = False

    def _update_data(self) -> None:
        """Writes the stored physics data to each phyics class insatnce"""
        self.mw.data = self.data.mw
    
    def _set_mesh(self, mesh: Mesh3D) -> None:
        """Set the current model mesh to a given mesh."""
        logger.trace(f'Setting {mesh} as model mesh')
        self.mesh = mesh
        self.mw.mesh = mesh
        self.display._mesh = mesh
    
    def _save_geometries(self) -> None:
        """Saves the current geometry state to the simulatin dataset
        """
        logger.trace('Storing geometries in data.sim')
        self.data.sim['geos'] = {geo.name: geo for geo in _GEOMANAGER.all_geometries()}
        self.data.sim['mesh'] = self.mesh
        self.data.sim.entries.append(self.data.sim.stock)
    ############################################################
    #                       PUBLIC FUNCTIONS                  #
    ############################################################

    def cache_build(self) -> bool:
        """Checks if all the lines inside this if statement block are the same as those
        stored from a previous run. If so, then it returns false. Else it returns True.
        
        Can be used to capture an entire model simulation.
        
        Example:
        
        >>> if model.cache_build():
        >>>     box = em.geo.Box(...)
        >>>     # Other lines
        >>>     model.mw.run_sweep()
        >>> data = model.data.mw

        Returns:
            bool: If the code is not the same
        """
        
        self.save_file = True
        self._cache_run = True
        filestr = get_build_section()
        self._file_lines = filestr
        cachepath = self.modelpath / 'pylines.txt'
        
        # If there is not pylines file, simulate (can't have been run).
        if not cachepath.exists():
            logger.info('No cached data detected, running file')
            return True
        
        with open(cachepath, 'r') as file:
            lines = file.read()
        
        if lines==filestr:
            logger.info('Cached data detected! Loading data!')
            self.load()
            return False
        logger.info('Different cached data detected, rebuilding file.')
        return True
    
    def cache_run(self) -> bool:
        """Checks if all the lines before this call are the same as the lines
        stored from a previous run. If so, then it returns false. Else it returns True.
        
        Can be used to capture a run_sweep() call.
        
        Example:
        
        >>> if model.cache_run():
        >>>     model.mw.run_sweep()
        >>> data = model.data.mw

        Returns:
            bool: If the code is not the same
        """
        self.save_file = True
        self._cache_run = True
        filestr = get_run_section()
        self._file_lines = filestr
        cachepath = self.modelpath / 'pylines.txt'
        
        # If there is not pylines file, simulate (can't have been run).
        if not cachepath.exists():
            logger.info('No cached data detected, running simulation!')
            return True
        
        with open(cachepath, 'r') as file:
            lines = file.read()
        
        if lines==filestr:
            logger.info('Cached data detected! Loading data!')
            self.load()
            return False
        logger.info('Different cached data detected, rerunning simulation.')
        return True
    
    def check_version(self, target_version: str, *, log: bool = False) -> None:
        """
        Ensure the script targets an EMerge version compatible with the current runtime.

        Parameters
        ----------
        target_version : str
            The EMerge version this script was written for (e.g. "1.4.0").
        log : bool, optional
            If True and a `logger` is available, emit a single WARNING with the same
            message as the exception. Defaults to False.

        Raises
        ------
        VersionError
            If the script's target version differs from the running EMerge version.
        """
        try:
            from packaging.version import Version as _V
            v_script = _V(target_version)
            v_runtime = _V(__version__)
            newer = v_script > v_runtime
            older = v_script < v_runtime
        except Exception:
            def _parse(v: str):
                try:
                    return tuple(int(p) for p in v.split("."))
                except Exception:
                    # Last-resort: compare as strings to avoid crashing the check itself
                    return tuple(v.split("."))
            v_script = _parse(target_version)
            v_runtime = _parse(__version__)
            newer = v_script > v_runtime
            older = v_script < v_runtime

        if not newer and not older:
            return  # exact match

        if newer:
            msg = (
                f"Script targets EMerge {target_version}, but runtime is {__version__}. "
                "The script may rely on features added after your installed version. "
                "Recommended: upgrade EMerge (`pip install --upgrade emerge`). "
                "If you know the script is compatible, you may remove this check."
            )
        else:  # older
            msg = (
                f"Script targets EMerge {target_version}, but runtime is {__version__}. "
                "APIs may have changed since the targeted version. "
                "Recommended: update the script for the current EMerge, or run a matching older release. "
                "If you know the script is compatible, you may remove this check."
            )

        if log:
            try:
                logger.warning(msg)
            except Exception:
                pass

        raise VersionError(msg)

    def save(self) -> None:
        """Saves the current model in the provided project directory."""
        # Ensure directory exists
        if not self.modelpath.exists():
            self.modelpath.mkdir(parents=True, exist_ok=True)
            logger.info(f"Created directory: {self.modelpath}")

        # Save mesh
        mesh_path = self.modelpath / 'mesh.msh'
        brep_path = self.modelpath / 'model.brep'

        gmsh.option.setNumber('Mesh.SaveParametric', 1)
        gmsh.option.setNumber('Mesh.SaveAll', 1)
        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()

        gmsh.write(str(mesh_path))
        gmsh.write(str(brep_path))
        logger.info(f"Saved mesh to: {mesh_path}")

        # Pack and save data
        dataset = dict(simdata=self.data, mesh=self.mesh)
        data_path = self.modelpath / 'simdata.emerge'
        
        joblib.dump(dataset, str(data_path))
        
        if self._cache_run:
            cachepath = self.modelpath / 'pylines.txt'
            with open(str(cachepath), 'w') as f_out:
                f_out.write(self._file_lines)
            
        logger.info(f"Saved simulation data to: {data_path}")

    def load(self) -> None:
        """Loads the model from the project directory."""
        mesh_path = self.modelpath / 'mesh.msh'
        brep_path = self.modelpath / 'model.brep'
        data_path = self.modelpath / 'simdata.emerge'

        if not mesh_path.exists() or not data_path.exists():
            raise FileNotFoundError("Missing required mesh or data file.")

        # Load mesh
        gmsh.open(str(brep_path))
        gmsh.merge(str(mesh_path))
        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()
        logger.info(f"Loaded mesh from: {mesh_path}")
        
        datapack = joblib.load(str(data_path))
        
        self.data = datapack['simdata']
        self.activate(0)
        
        logger.info(f"Loaded simulation data from: {data_path}")
    
    def set_loglevel(self, loglevel: Literal['DEBUG','INFO','WARNING','ERROR']) -> None:
        """Set the loglevel for loguru.

        Args:
            loglevel ('DEBUG','INFO','WARNING','ERROR'): The loglevel
        """
        logger.trace(f'Setting loglevel to {loglevel}')
        LOG_CONTROLLER.set_std_loglevel(loglevel)
        if loglevel not in ('TRACE','DEBUG'):
            gmsh.option.setNumber("General.Terminal", 0)

    def set_write_log(self) -> None:
        """Adds a file output for the logger."""
        logger.trace(f'Writing log to path = {self.modelpath}')
        LOG_CONTROLLER.set_write_file(self.modelpath)
        
    def view(self, 
             selections: list[Selection] | None = None, 
             use_gmsh: bool = False,
             plot_mesh: bool = False,
             volume_mesh: bool = True,
             opacity: float | None = None,
             labels: bool = False) -> None:
        """View the current geometry in either the BaseDisplay object (PVDisplay only) or
        the GMSH viewer.

        Args:
            selections (list[Selection] | None, optional): Optional selections to highlight. Defaults to None.
            use_gmsh (bool, optional): If GMSH's GUI should be used. Defaults to False.
            plot_mesh (bool, optional): If the mesh should be plot instead of the object. Defaults to False.
            volume_mesh (bool, optional): If the internal mesh should be plot instead of only the surface boundary mesh. Defaults to True
            opacity (float | None, optional): The object/mesh opacity. Defaults to None.
            labels: (bool, optional): If geometry name labels should be shown. Defaults to False.
        """
        if not (self.display is not None and self.mesh.defined) or use_gmsh:
            gmsh.model.occ.synchronize()
            gmsh.fltk.run()
            return
        for geo in _GEOMANAGER.all_geometries():
            self.display.add_object(geo, mesh=plot_mesh, opacity=opacity, volume_mesh=volume_mesh, label=labels)
        if selections:
            [self.display.add_object(sel, color='red', opacity=0.6, label=labels) for sel in selections]
        self.display.show()

        return None
        
    def set_periodic_cell(self, cell: PeriodicCell):
        """Set the given periodic cell object as the simulations peridicity.

        Args:
            cell (PeriodicCell): The PeriodicCell class
        """
        logger.trace(f'Setting {cell} as periodic cell object')
        self.mw.bc._cell = cell
        self._cell = cell

    def set_resolution(self, resolution: float) -> Simulation:
        """Sets the discretization resolution in the various physics interfaces.
        

        Args:
            resolution (float): The resolution as a float. Lower resolution is a finer mesh 

        Returns:
            Simulation: _description_
        """
        self.mw.set_resolution(resolution)
        
    def commit_geometry(self, *geometries: GeoObject | list[GeoObject]) -> None:
        """Finalizes and locks the current geometry state of the simulation.

        The geometries may be provided (legacy behavior) but are automatically managed in the background.
        
        """
        geometries_parsed: Any = None
        logger.trace('Committing final geometry.')
        if not geometries:
            geometries_parsed = _GEOMANAGER.all_geometries()
        else:
            geometries_parsed = unpack_lists(geometries + tuple([item for item in self.data.sim.default.values() if isinstance(item, GeoObject)]))
        logger.trace(f'Parsed geometries = {geometries_parsed}')
        
        self._save_geometries()
        
        self.mesher.submit_objects(geometries_parsed)
        self._defined_geometries = True
        self.display._facetags = [dt[1] for dt in gmsh.model.get_entities(2)]
    
    def all_geos(self) -> list[GeoObject]:
        """Returns all geometries in a list

        Returns:
            list[GeoObject]: A list of all GeoObjects
        """
        return _GEOMANAGER.all_geometries()  
    
    def activate(self, _indx: int | None = None, **variables) -> Simulation:
        """Searches for the permutaions of parameter sweep variables and sets the current geometry to the provided set."""
        if _indx is not None:
            dataset = self.data.sim.index(_indx)
        else:
            dataset = self.data.sim.find(**variables)
        
        variables = ', '.join([f'{key}={value}' for key,value in dataset.vars.items()])
        logger.info(f'Activated entry with variables: {variables}')
        _GEOMANAGER.set_geometries(dataset['geos'])
        self._set_mesh(dataset['mesh'])
        return self
        
    def generate_mesh(self) -> None:
        """Generate the mesh. 
        This can only be done after commit_geometry(...) is called and if frequencies are defined.

        Args:
            name (str, optional): The mesh file name. Defaults to "meshname.msh".

        Raises:
            ValueError: ValueError if no frequencies are defined.
        """
        logger.trace('Starting mesh generation phase.')
        if not self._defined_geometries:
            self.commit_geometry()
        
        logger.trace(' (1) Installing periodic boundaries in mesher.')
        # Set the cell periodicity in GMSH
        if self._cell is not None:
            self.mesher.set_periodic_cell(self._cell)
        
        self.mw._initialize_bcs(_GEOMANAGER.get_surfaces())

        # Check if frequencies are defined: TODO: Replace with a more generic check
        if self.mw.frequencies is None:
            raise ValueError('No frequencies defined for the simulation. Please set frequencies before generating the mesh.')

        gmsh.model.occ.synchronize()

        # Set the mesh size
        self.mesher._configure_mesh_size(self.mw.get_discretizer(), self.mw.resolution)
        
        logger.trace(' (2) Calling GMSH mesher')
        try:
            gmsh.logger.start()
            gmsh.model.mesh.generate(3)
            logs = gmsh.logger.get()
            gmsh.logger.stop()
            for log in logs:
                logger.trace('[GMSH] ' + log)
        except Exception:
            logger.error('GMSH Mesh error detected.')
            print(_GMSH_ERROR_TEXT)
            raise
        logger.info('GMSH Meshing complete!')
        self.mesh._pre_update(self.mesher._get_periodic_bcs())
        self.mesh.exterior_face_tags = self.mesher.domain_boundary_face_tags
        gmsh.model.occ.synchronize()
        self._set_mesh(self.mesh)
        logger.trace(' (3) Mesh routine complete')

    def parameter_sweep(self, clear_mesh: bool = True, **parameters: np.ndarray) -> Generator[tuple[float,...], None, None]:
        """Executes a parameteric sweep iteration.

        The first argument clear_mesh determines if the mesh should be cleared and rebuild in between sweeps. This is usually needed
        except when you change only port-properties or material properties. The parameters of the sweep can be provided as a set of 
        keyword arguments. As an example if you defined the axes: width=np.linspace(...) and height=np.linspace(...). You can
        add them as arguments using .parameter_sweep(True, width=width, height=height).

        The rest of the simulation commands should be inside the iteration scope

        Args:
            clear_mesh (bool, optional): Wether to clear the mesh in between sweeps. Defaults to True.

        Yields:
            Generator[tuple[float,...], None, None]: The parameters provided

        Example:
         >>> for W, H in model.parameter_sweep(True, width=widths, height=heights):
         >>>    // build simulation
         >>>    data = model.run_sweep()
         >>> // Extract the data
         >>> widths, heights, frequencies, S21 = data.ax('width','height','freq').S(2,1)
        """
        paramlist = sorted(list(parameters.keys()))
        dims = np.meshgrid(*[parameters[key] for key in paramlist], indexing='ij')
        dims_flat = [dim.flatten() for dim in dims]
        
        self.mw.cache_matrices = False
        logger.trace('Starting parameter sweep.')
        
        for i_iter in range(dims_flat[0].shape[0]):
            
            if clear_mesh and i_iter > 0:
                logger.info('Cleaning up mesh.')
                gmsh.clear()
                mesh = Mesh3D(self.mesher)
                _GEOMANAGER.reset(self.modelname)
                self._set_mesh(mesh)
                self.mw.reset()
            
            params = {key: dim[i_iter] for key,dim in zip(paramlist, dims_flat)}
            self.mw._params = params
            self.data.sim.new(**params)
            
            logger.info(f'Iterating: {params}')
            if len(dims_flat)==1:
                yield dims_flat[0][i_iter]
            else:
                yield (dim[i_iter] for dim in dims_flat) # type: ignore
            
            if not clear_mesh:
                self._save_geometries()
        
        if not clear_mesh:
                self._save_geometries()
        
        self.mw.cache_matrices = True
    
    def export(self, filename: str):
        """Exports the model or mesh depending on the extension. 
        
        Exporting is realized by GMSH.
        Supported file formats are:
        
        3D Model: .opt, .geo_unrolled, .brep, .xao ,.step and .iges
        Mesh: .msh, .inp, .key, ._0000.rad, .celum, .cgns, .diff, .unv, .ir3, .mes, .mesh 
              .mail, .m, .bdf, .off, .p3d, .stl, .wrl, .vtk, .dat, .ply2, .su2, .neu, .x3d

        Args:
            filename (str): The filename
        """
        logger.trace(f'Writing geometry to {filename}')
        gmsh.write(filename)
        
    def set_solver(self, solver: EMSolver | Solver):
        """Set a given Solver class instance as the main solver. 
        Solvers will be checked on validity for the given problem.

        Args:
            solver (EMSolver | Solver): The solver objects
        """
        logger.trace(f'Setting solver to {solver}')
        self.mw.solveroutine.set_solver(solver)
        
    ############################################################
    #                     DEPRICATED FUNCTIONS                #
    ############################################################

    def define_geometry(self, *args):
        """DEPRICATED VERSION: Use .commit_geometry()
        """
        logger.warning('define_geometry() will be derpicated. Use commit_geometry() instead.')
        self.commit_geometry(*args)
        


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.


import numpy as np
from ....elements import Nedelec2
from ....mth.optimized import local_mapping, compute_distances
from numba import c16, types, f8, i8, njit, prange


############################################################
#                      FIELD MAPPING                      #
############################################################

@njit(i8[:,:](i8, i8[:,:], i8[:,:], i8[:,:]), cache=True, nogil=True)
def local_tri_to_edgeid(itri: int, tris, edges, tri_to_edge) -> np.ndarray:
    global_edge_map = edges[:, tri_to_edge[:,itri]]
    return local_mapping(tris[:, itri], global_edge_map)


@njit(cache=True, fastmath=True, nogil=True)
def optim_matmul(B: np.ndarray, data: np.ndarray):
    dnew = np.zeros_like(data)
    dnew[0,:] = B[0,0]*data[0,:] + B[0,1]*data[1,:] + B[0,2]*data[2,:]
    dnew[1,:] = B[1,0]*data[0,:] + B[1,1]*data[1,:] + B[1,2]*data[2,:]
    dnew[2,:] = B[2,0]*data[0,:] + B[2,1]*data[1,:] + B[2,2]*data[2,:]
    return dnew

@njit(f8[:](f8[:], f8[:]), cache=True, fastmath=True, nogil=True)
def cross(a: np.ndarray, b: np.ndarray):
    crossv = np.empty((3,), dtype=np.float64)
    crossv[0] = a[1]*b[2] - a[2]*b[1]
    crossv[1] = a[2]*b[0] - a[0]*b[2]
    crossv[2] = a[0]*b[1] - a[1]*b[0]
    return crossv

@njit(cache=True, nogil=True)
def normalize(a: np.ndarray):
    return a/((a[0]**2 + a[1]**2 + a[2]**2)**0.5)


############################################################
#              GAUSS QUADRATURE IMPLEMENTATION             #
############################################################

@njit(c16(c16[:], c16[:], types.Array(types.float64, 1, 'A', readonly=True)), cache=True, nogil=True)
def _gqi(v1, v2, W):
    return np.sum(v1*v2*W)


############################################################
#                BASIS FUNCTION DERIVATIVES               #
############################################################


@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _curl_edge_1(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    out = -3*a1*b1*c2 + 3*a1*b2*c1 - 3*b1**2*c2*xs + 3*b1*b2*c1*xs - 3*b1*c1*c2*ys + 3*b2*c1**2*ys + 0j
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _curl_edge_2(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    out = -3*a2*b1*c2 + 3*a2*b2*c1 - 3*b1*b2*c2*xs - 3*b1*c2**2*ys + 3*b2**2*c1*xs + 3*b2*c1*c2*ys + 0j
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _curl_face_1(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    out = -b2*(c1*(a3 + b3*xs + c3*ys) - c3*(a1 + b1*xs + c1*ys)) + c2*(b1*(a3 + b3*xs + c3*ys) - b3*(a1 + b1*xs + c1*ys)) + 2*(b1*c3 - b3*c1)*(a2 + b2*xs + c2*ys) + 0j
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _curl_face_2(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    out = b3*(c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys)) - c3*(b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys)) - 2*(b1*c2 - b2*c1)*(a3 + b3*xs + c3*ys) + 0j
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _divergence_edge_1(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    out = b1*(b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys)) + c1*(c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys)) + 0j
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _divergence_edge_2(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    out = b2*(b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys)) + c2*(c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys)) + 0j
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _divergence_face_1(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    out = -b2*(b1*(a3 + b3*xs + c3*ys) - b3*(a1 + b1*xs + c1*ys)) - c2*(c1*(a3 + b3*xs + c3*ys) - c3*(a1 + b1*xs + c1*ys)) + 0j
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True)
def _divergence_face_2(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    out = b3*(b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys)) + c3*(c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys)) + 0j
    return out

############################################################
#     TRIANGLE BARYCENTRIC COORDINATE LIN. COEFFICIENTS    #
############################################################


@njit(types.Tuple((f8[:], f8[:], f8[:], f8))(f8[:], f8[:]), cache = True, nogil=True)
def tri_coefficients(vxs, vys):

    x1, x2, x3 = vxs
    y1, y2, y3 = vys

    a1 = x2*y3-y2*x3
    a2 = x3*y1-y3*x1
    a3 = x1*y2-y1*x2
    b1 = y2-y3
    b2 = y3-y1
    b3 = y1-y2
    c1 = x3-x2
    c2 = x1-x3
    c3 = x2-x1

    sA = 0.5*(((x1-x3)*(y2-y1) - (x1-x2)*(y3-y1)))
    sign = np.sign(sA)
    A = np.abs(sA)
    As = np.array([a1, a2, a3])*sign
    Bs = np.array([b1, b2, b3])*sign
    Cs = np.array([c1, c2, c3])*sign
    return As, Bs, Cs, A


############################################################
#                  GAUSS QUADRATURE POINTS                 #
############################################################


DPTS = np.array([[0.22338159, 0.22338159, 0.22338159, 0.10995174, 0.10995174, 0.10995174],
                 [0.10810302, 0.44594849, 0.44594849, 0.81684757, 0.09157621, 0.09157621],
                 [0.44594849, 0.44594849, 0.10810302, 0.09157621, 0.09157621, 0.81684757],
                 [0.44594849, 0.10810302, 0.44594849, 0.09157621, 0.81684757, 0.09157621]], dtype=np.float64)


############################################################
#                 NUMBA OPTIMIZED ASSEMBLER                #
############################################################


@njit(c16[:,:](f8[:,:], i8[:,:], c16), cache=True, nogil=True)
def _abc_order_2_terms(tri_vertices, local_edge_map, cf):
    '''ABC order 2 tangent gradient term'''
    
    origin = tri_vertices[:,0]
    vertex_2 = tri_vertices[:,1]
    vertex_3 = tri_vertices[:,2]
    
    edge_1 = vertex_2-origin
    edge_2 = vertex_3-origin
    
    zhat = normalize(cross(edge_1, edge_2))
    xhat = normalize(edge_1)
    yhat = normalize(cross(zhat, xhat))
    
    basis = np.zeros((3,3), dtype=np.float64)
    basis[0,:] = xhat
    basis[1,:] = yhat
    basis[2,:] = zhat
    
    local_vertices = optim_matmul(basis, tri_vertices - origin[:,np.newaxis])
    
    CurlMatrix = np.zeros((8,8), dtype=np.complex128)
    DivMatrix = np.zeros((8,8), dtype=np.complex128)
    
    Lengths = np.ones((8,8), dtype=np.float64)

    WEIGHTS = DPTS[0,:]
    DPTS1 = DPTS[1,:]
    DPTS2 = DPTS[2,:]
    DPTS3 = DPTS[3,:]

    xpts = local_vertices[0,:]
    ypts = local_vertices[1,:]

    distances = compute_distances(xpts, ypts, 0*xpts)

    xs = xpts[0]*DPTS1 + xpts[1]*DPTS2 + xpts[2]*DPTS3
    ys = ypts[0]*DPTS1 + ypts[1]*DPTS2 + ypts[2]*DPTS3
    
    int_coords = np.empty((2,xs.shape[0]), dtype=np.float64)
    int_coords[0,:] = xs
    int_coords[1,:] = ys

    aas, bbs, ccs, Area = tri_coefficients(xpts, ypts)
    
    bary_coeff = np.empty((3,3), dtype=np.float64)
    bary_coeff[0,:] = aas/(2*Area)
    bary_coeff[1,:] = bbs/(2*Area)
    bary_coeff[2,:] = ccs/(2*Area)

    Lengths[3,:] *= distances[0,2]
    Lengths[7,:] *= distances[0,1]
    Lengths[:,3] *= distances[0,2]
    Lengths[:,7] *= distances[0,1]
    
    FF1C = _curl_face_1(bary_coeff, int_coords)
    FF2C = _curl_face_2(bary_coeff, int_coords)
    FF1D = _divergence_face_1(bary_coeff, int_coords)
    FF2D = _divergence_face_2(bary_coeff, int_coords)
    
    for iv1 in range(3):
        ie1 = local_edge_map[:, iv1]
        
        Le = distances[ie1[0], ie1[1]]
        Lengths[iv1,:] *= Le
        Lengths[iv1+4,:] *= Le
        Lengths[:,iv1] *= Le
        Lengths[:,iv1+4] *= Le
        
        FE1C_1 = _curl_edge_1(bary_coeff[:,ie1], int_coords)
        FE2C_1 = _curl_edge_2(bary_coeff[:,ie1], int_coords)
        FE1D_1 = _divergence_edge_1(bary_coeff[:,ie1], int_coords)
        FE2D_1 = _divergence_edge_2(bary_coeff[:,ie1], int_coords)
        
        for iv2 in range(3):
            ie2 = local_edge_map[:, iv2]
            
            FE1C_2 = _curl_edge_1(bary_coeff[:,ie2], int_coords)
            FE2C_2 = _curl_edge_2(bary_coeff[:,ie2], int_coords)
            FE1D_2 = _divergence_edge_1(bary_coeff[:,ie2], int_coords)
            FE2D_2 = _divergence_edge_2(bary_coeff[:,ie2], int_coords)
            
            CurlMatrix[iv1, iv2]     = _gqi(FE1C_1, FE1C_2, WEIGHTS)
            CurlMatrix[iv1, iv2+4]   = _gqi(FE1C_1, FE2C_2, WEIGHTS)
            CurlMatrix[iv1+4, iv2]   = _gqi(FE2C_1, FE1C_2, WEIGHTS)
            CurlMatrix[iv1+4, iv2+4] = _gqi(FE2C_1, FE2C_2, WEIGHTS)
            
            DivMatrix[iv1, iv2]     = _gqi(FE1D_1, FE1D_2, WEIGHTS)
            DivMatrix[iv1, iv2+4]   = _gqi(FE1D_1, FE2D_2, WEIGHTS)
            DivMatrix[iv1+4, iv2]   = _gqi(FE2D_1, FE1D_2, WEIGHTS)
            DivMatrix[iv1+4, iv2+4] = _gqi(FE2D_1, FE2D_2, WEIGHTS)
        
        CurlMatrix[iv1,  3]      = _gqi(FE1C_1, FF1C, WEIGHTS)
        CurlMatrix[iv1+4,3]      = _gqi(FE2C_1, FF1C, WEIGHTS)
        CurlMatrix[iv1,  7]      = _gqi(FE1C_1, FF2C, WEIGHTS)
        CurlMatrix[iv1+4,7]      = _gqi(FE2C_1, FF2C, WEIGHTS)
        
        CurlMatrix[3, iv1]   = CurlMatrix[iv1, 3]
        CurlMatrix[3, iv1+4] = CurlMatrix[iv1+4, 3]
        CurlMatrix[7, iv1]   = CurlMatrix[iv1, 7]
        CurlMatrix[7, iv1+4] = CurlMatrix[iv1+4, 7]
        
        DivMatrix[iv1,  3]      = _gqi(FE1D_1, FF1D, WEIGHTS)
        DivMatrix[iv1+4,3]      = _gqi(FE2D_1, FF1D, WEIGHTS)
        DivMatrix[iv1,  7]      = _gqi(FE1D_1, FF2D, WEIGHTS)
        DivMatrix[iv1+4,7]      = _gqi(FE2D_1, FF2D, WEIGHTS)
        
        DivMatrix[3, iv1]   = DivMatrix[iv1, 3]
        DivMatrix[3, iv1+4] = DivMatrix[iv1+4, 3]
        DivMatrix[7, iv1]   = DivMatrix[iv1, 7]
        DivMatrix[7, iv1+4] = DivMatrix[iv1+4, 7]
    
    CurlMatrix[3, 3] = _gqi(FF1C, FF1C, WEIGHTS)
    CurlMatrix[3, 7] = _gqi(FF1C, FF2C, WEIGHTS)
    CurlMatrix[7, 3] = _gqi(FF2C, FF1C, WEIGHTS)
    CurlMatrix[7, 7] = _gqi(FF2C, FF2C, WEIGHTS)
    
    DivMatrix[3, 3] = _gqi(FF1D, FF1D, WEIGHTS)
    DivMatrix[3, 7] = _gqi(FF1D, FF2D, WEIGHTS)
    DivMatrix[7, 3] = _gqi(FF2D, FF1D, WEIGHTS)
    DivMatrix[7, 7] = _gqi(FF2D, FF2D, WEIGHTS)
    
    Mat = cf*Lengths*(CurlMatrix-DivMatrix)*np.abs(Area)
    return Mat


############################################################
#            NUMBA OPTIMIZED INTEGRAL OVER TETS            #
############################################################

@njit((c16[:])(f8[:,:],
            i8[:,:], 
            i8[:,:], 
            i8[:,:],
            i8[:],
            c16), cache=True, nogil=True, parallel=True)
def _matrix_builder(nodes, tris, edges, tri_to_field, tri_ids, coeff):
    """ Numba optimized loop over each face triangle."""
    ntritot = tris.shape[1]
    nnz = ntritot*64
    
    Mat = np.zeros(nnz, dtype=np.complex128)

    tri_to_edge = tri_to_field[:3,:]
    
    Ntris = tri_ids.shape[0]
    for itri_sub in prange(Ntris): # type: ignore
        
        itri = tri_ids[itri_sub]
        p = itri*64

        # Construct a local mapping to global triangle orientations
        local_tri_map = local_tri_to_edgeid(itri, tris, edges, tri_to_edge)

        # Construct the local edge map
        tri_nodes = nodes[:, tris[:,itri]]
        subMat = _abc_order_2_terms(tri_nodes, local_tri_map, coeff)

        Mat[p:p+64] += subMat.ravel()
        
    return Mat

############################################################
#                     PYTHON INTERFACE                     #
############################################################

    
def abc_order_2_matrix(field: Nedelec2,
                       surf_triangle_indices: np.ndarray,
                       coeff: complex) -> np.ndarray:
    """Computes the second order absorbing boundary condition correction terms.

    Args:
        field (Nedelec2): The Basis function object
        surf_triangle_indices (np.ndarray): The surface triangle indices to add
        coeff (complex): The integral coefficient jp2/k0

    Returns:
        np.ndarray: The resultant matrix items
    """
    Mat = _matrix_builder(field.mesh.nodes, field.mesh.tris, field.mesh.edges, field.tri_to_field, surf_triangle_indices, coeff)
    return Mat


import subprocess
import sys
from importlib import import_module

try:
    from loguru import logger
except ImportError:
    # Fallback if loguru isn't installed
    class _Logger:
        def info(self, msg): print(f"[INFO] {msg}")
        def warning(self, msg): print(f"[WARN] {msg}")
        def error(self, msg): print(f"[ERROR] {msg}")
    logger = _Logger()


def update_emerge(branch: str = "main", confirm: bool = True, _dryrun: bool = False) -> str | None:
    """
    Update the EMerge library directly from GitHub.

    Parameters
    ----------
    branch : str, optional
        The git branch or tag to install from. Default is "main".
    confirm : bool, optional
        If True, asks the user for confirmation before updating.

    Returns
    -------
    str | None
        The updated version string if successful, otherwise None.
    """
    logger.warning(
        "You are about to update EMerge from GitHub. "
        "This is an experimental feature and may overwrite your installed version."
    )

    if confirm:
        ans = input("Do you wish to proceed? [y/N] ").strip().lower()
        if ans != "y":
            logger.info("Update aborted by user.")
            return None

    url = f"git+https://github.com/FennisRobert/EMerge.git@{branch}"
    logger.info(f"Updating EMerge from branch/tag '{branch}'...")

    if not _dryrun:
        try:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "--upgrade", "--force-reinstall", url]
            )
        except subprocess.CalledProcessError as e:
            logger.error(f"Update failed: {e}")
            return None
    else:
        logger.info('Dry run... No update executed!')
    try:
        emerge = import_module("emerge")
        version = getattr(emerge, "__version__", "unknown")
        logger.info(f"Successfully updated to EMerge version {version}.")
        return version
    except Exception as e:
        logger.error(f"Update installed but could not retrieve version: {e}")
        return None